r"""
===================================
Intravoxel incoherent motion (IVIM)
===================================
The intravoxel incoherent motion (IVIM) model describes diffusion
and perfusion in the signal acquired with a diffusion MRI sequence
that contains multiple low b-values. The IVIM model can be understood
as an adaptation of the work of Stejskal and Tanner :footcite:p:`Stejskal1965`
in biological tissue, and was proposed by Le Bihan :footcite:p:`LeBihan1988`.
The model assumes two compartments: a slow moving compartment,
where particles diffuse in a Brownian fashion as a consequence of thermal
energy, and a fast moving compartment (the vascular compartment), where
blood moves as a consequence of a pressure gradient. In the first compartment,
the diffusion coefficient is $\mathbf{D}$ while in the second compartment, a
pseudo diffusion term $\mathbf{D^*}$ is introduced that describes the
displacement of the blood elements in an assumed randomly laid out vascular
network, at the macroscopic level. According to :footcite:p:`LeBihan1988`,
$\mathbf{D^*}$ is greater than $\mathbf{D}$.

The IVIM model expresses the MRI signal as follows:

 .. math::
    S(b)=S_0(fe^{-bD^*}+(1-f)e^{-bD})

where $\mathbf{b}$ is the diffusion gradient weighing value (which is dependent
on the measurement parameters), $\mathbf{S_{0}}$ is the signal in the absence
of diffusion gradient sensitization, $\mathbf{f}$ is the perfusion
fraction, $\mathbf{D}$ is the diffusion coefficient and $\mathbf{D^*}$ is
the pseudo-diffusion constant, due to vascular contributions.

In the following example we show how to fit the IVIM model on a
diffusion-weighted dataset and visualize the diffusion and pseudo-diffusion
coefficients. First, we import all relevant modules:
"""

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
from dipy.reconst.ivim import IvimModel

###############################################################################
# We get an IVIM dataset using DIPY_'s data fetcher ``read_ivim``.
# This dataset was acquired with 21 b-values in 3 different directions.
# Volumes corresponding to different directions were registered to each
# other, and averaged across directions. Thus, this dataset has 4 dimensions,
# with the length of the last dimension corresponding to the number
# of b-values. In order to use this model the data should contain signals
# measured at 0 bvalue.

fraw, fbval, fbvec = get_fnames(name="ivim")

###############################################################################
# The gtab contains a GradientTable object (information about the gradients
# e.g. b-values and b-vectors). We get the data from the file using
# ``load_nifti_data``.

data = load_nifti_data(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs, b0_threshold=0)
print(f"data.shape {data.shape}")

###############################################################################
# The data has 54 slices, with 256-by-256 voxels in each slice. The fourth
# dimension corresponds to the b-values in the gtab. Let us visualize the data
# by taking a slice midway(z=33) at $\mathbf{b} = 0$.

z = 33
b = 0

plt.imshow(data[:, :, z, b].T, origin="lower", cmap="gray", interpolation="nearest")
plt.axhline(y=100)
plt.axvline(x=170)
plt.savefig("ivim_data_slice.png")
plt.close()

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Heat map of a slice of data
#
#
# The region around the intersection of the cross-hairs in the figure
# contains cerebral spinal fluid (CSF), so it should have a very high
# $\mathbf{f}$ and $\mathbf{D^*}$, the area just medial to that is white
# matter so that should be lower, and the region more laterally contains a
# mixture of gray matter and CSF. That should give us some contrast to see
# the values varying across the regions.

x1, x2 = 90, 155
y1, y2 = 90, 170
data_slice = data[x1:x2, y1:y2, z, :]

plt.imshow(
    data[x1:x2, y1:y2, z, b].T, origin="lower", cmap="gray", interpolation="nearest"
)
plt.savefig("CSF_slice.png")
plt.close()

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Heat map of the CSF slice selected.
#
#
# Now that we have prepared the datasets we can go forward with
# the ivim fit. We provide two methods of fitting the parameters of the IVIM
# multi-exponential model explained above. We first fit the model with a simple
# fitting approach by passing the option `fit_method='trr'`. This method uses
# a two-stage approach: first, a linear fit used to get quick initial guesses
# for the parameters $\mathbf{S_{0}}$ and $\mathbf{D}$ by considering b-values
# greater than ``split_b_D`` (default: 400))and assuming a mono-exponential
# signal. This is based on the assumption that at high b-values the signal can
# be approximated as a mono exponential decay and by taking the logarithm of
# the signal values a linear fit can be obtained. Another linear fit for ``S0``
# (bvals < ``split_b_S0`` (default: 200)) follows and ``f`` is estimated using
# $1 - S0_{prime}/S0$. Then a non-linear least-squares fitting is performed to
# fit ``D_star`` and ``f``. If the ``two_stage`` flag is set to ``True`` while
# initializing the model, a final non-linear least squares fitting is performed
# for all the parameters. All initializations for the model such as
# ``split_b_D`` are passed while creating the ``IvimModel``. If you are
# using Scipy 0.17, you can also set bounds by setting
# ``bounds=([0., 0., 0.,0.], [np.inf, 1., 1., 1.]))`` while initializing the
# ``IvimModel``.
#
# For brevity, we focus on a small section of the slice as selected above,
# to fit the IVIM model. First, we instantiate the IvimModel object.

ivimmodel = IvimModel(gtab, fit_method="trr")

###############################################################################
# To fit the model, call the `fit` method and pass the data for fitting.

ivimfit = ivimmodel.fit(data_slice)

###############################################################################
# The fit method creates a IvimFit object which contains the
# parameters of the model obtained after fitting. These are accessible
# through the `model_params` attribute of the IvimFit object.
# The parameters are arranged as a 4D array, corresponding to the spatial
# dimensions of the data, and the last dimension (of length 4)
# corresponding to the model parameters according to the following
# order : $\mathbf{S_{0}, f, D^*, D}$.

ivimparams = ivimfit.model_params
print(f"ivimparams.shape : {ivimparams.shape}")

###############################################################################
# As we see, we have a 20x20 slice at the height z = 33. Thus we
# have 400 voxels. We will now plot the parameters obtained from the
# fit for a voxel and also various maps for the entire slice.
# This will give us an idea about the diffusion and perfusion in
# that section. Let(i, j) denote the coordinate of the voxel. We have
# already fixed the z component as 33 and hence we will get a slice
# which is 33 units above.

i, j = 10, 10
estimated_params = ivimfit.model_params[i, j, :]
print(estimated_params)

###############################################################################
# Now we can map the perfusion and diffusion maps for the slice. We
# will plot a heatmap showing the values using a colormap. It will be
# useful to define a plotting function for the heatmap here since we
# will use it to plot for all the IVIM parameters. We will need to specify
# the lower and upper limits for our data. For example, the perfusion
# fractions should be in the range (0,1). Similarly, the diffusion and
# pseudo-diffusion constants are much smaller than 1. We pass an argument
# called ``variable`` to out plotting function which gives the label for
# the plot.


def plot_map(raw_data, variable, limits, filename):
    fig, ax = plt.subplots(1)
    lower, upper = limits
    ax.set_title(f"Map for {variable}")
    im = ax.imshow(
        raw_data.T,
        origin="lower",
        clim=(lower, upper),
        cmap="gray",
        interpolation="nearest",
    )
    fig.colorbar(im)
    fig.savefig(filename)


###############################################################################
# Let us get the various plots with `fit_method = 'trr'` so that we can
# visualize them in one page


plot_map(ivimfit.S0_predicted, "Predicted S0", (0, 10000), "predicted_S0.png")
plot_map(data_slice[:, :, 0], "Measured S0", (0, 10000), "measured_S0.png")
plot_map(ivimfit.perfusion_fraction, "f", (0, 1), "perfusion_fraction.png")
plot_map(ivimfit.D_star, "D*", (0, 0.01), "perfusion_coeff.png")
plot_map(ivimfit.D, "D", (0, 0.001), "diffusion_coeff.png")

###############################################################################
# Next, we will fit the same model with a more refined optimization process
# with `fit_method='VarPro'` (for "Variable Projection"). The VarPro computes
# the IVIM parameters using the MIX approach :footcite:p:`Farooq2016`. This
# algorithm uses three different optimizers. It starts with a differential
# evolution algorithm and fits the parameters in the power of exponentials. Then
# the fitted parameters in the first step are utilized to make a linear convex
# problem. Using a convex optimization, the volume fractions are determined.
# The last step is non-linear least-squares fitting on all the parameters.
# The results of the first and second optimizers are utilized as the initial
# values for the last step of the algorithm.
#
# As opposed to the `'trr'` fitting method, this approach does not need to set
# any thresholds on the bvals to differentiate between the perfusion
# (pseudo-diffusion) and diffusion portions and fits the parameters
# simultaneously. Making use of the three step optimization mentioned above
# increases the convergence basin for fitting the multi-exponential functions
# of microstructure models. This method has been described in further detail
# in :footcite:p:`Fadnavis2019` and :footcite:p:`Farooq2016`.

ivimmodel_vp = IvimModel(gtab, fit_method="VarPro")
ivimfit_vp = ivimmodel_vp.fit(data_slice)

###############################################################################
# Just like the `'trr'` fit method, `'VarPro'` creates a IvimFit object which
# contains the parameters of the model obtained after fitting. These are
# accessible through the `model_params` attribute of the IvimFit object.
# The parameters are arranged as a 4D array, corresponding to the spatial
# dimensions of the data, and the last dimension (of length 4)
# corresponding to the model parameters according to the following
# order : $\mathbf{S_{0}, f, D^*, D}$.

i, j = 10, 10
estimated_params = ivimfit_vp.model_params[i, j, :]
print(estimated_params)

###############################################################################
# To compare the fit using `fit_method='VarPro'` and `fit_method='trr'`, we can
# plot one voxel's signal and the model fit using both methods.
#
# We will use the `predict` method of the IvimFit objects, to get the predicted
# signal, based on each one of the model fit methods.

fig, ax = plt.subplots(1)

ax.scatter(gtab.bvals, data_slice[i, j, :], color="green", label="Measured signal")

ivim_trr_predict = ivimfit.predict(gtab)[i, j, :]

ax.plot(gtab.bvals, ivim_trr_predict, label="trr prediction")

S0_est, f_est, D_star_est, D_est = ivimfit.model_params[i, j, :]

trr_pro_param_est = (
    f"S0={S0_est:06.3f} f={f_est:06.4f}\n" f"D*={D_star_est:06.5f} D={D_est:06.5f}"
)
text_fit = f"""trr param estimates:\n {trr_pro_param_est}"""

ax.text(
    0.65,
    0.80,
    text_fit,
    horizontalalignment="center",
    verticalalignment="center",
    transform=plt.gca().transAxes,
)

ivim_predict_vp = ivimfit_vp.predict(gtab)[i, j, :]
ax.plot(gtab.bvals, ivim_predict_vp, label="VarPro prediction")

ax.set_xlabel("bvalues")
ax.set_ylabel("Signals")

S0_est, f_est, D_star_est, D_est = ivimfit_vp.model_params[i, j, :]

var_pro_param_est = (
    f"S0={S0_est:06.3f} f={f_est:06.4f}\n" f"D*={D_star_est:06.5f} D={D_est:06.5f}"
)
text_fit = f"""VarPro param estimates:\n{var_pro_param_est}"""

ax.text(
    0.65,
    0.50,
    text_fit,
    horizontalalignment="center",
    verticalalignment="center",
    transform=plt.gca().transAxes,
)

fig.legend(loc="upper right")
fig.savefig("ivim_voxel_plot.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Plot of the signal from one voxel.
#
#
# Let us get the various plots with `fit_method = 'VarPro'` so that we can
# visualize them in one page

plt.figure()
plot_map(
    ivimfit_vp.S0_predicted,
    "Heatmap of S0 predicted from the fit",
    (0, 10000),
    "predicted_S0.png",
)
plot_map(
    data_slice[..., 0],
    "Heatmap of measured signal at bvalue = 0",
    (0, 10000),
    "measured_S0.png",
)
plot_map(
    ivimfit_vp.perfusion_fraction,
    "Heatmap of perfusion fraction values predicted from the fit",
    (0, 1),
    "perfusion_fraction.png",
)
plot_map(
    ivimfit_vp.D_star,
    "D* - Heatmap of perfusion coefficients predicted from the fit",
    (0, 0.01),
    "perfusion_coeff.png",
)
plot_map(
    ivimfit_vp.D,
    "D - Heatmap of diffusion coefficients predicted from the fit",
    (0, 0.001),
    "diffusion_coeff.png",
)


###############################################################################
# References
# ----------
#
# .. footbibliography::
#



import os

import h5py
import numpy as np

from dipy.core.sphere import Sphere
from dipy.direction.peaks import PeaksAndMetrics, reshape_peaks_for_visualization
from dipy.io.image import save_nifti
from dipy.reconst.dti import quantize_evecs
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.deprecator import deprecate_with_version


def _safe_save(group, array, name):
    """Safe saving of arrays with specific names.

    Parameters
    ----------
    group : HDF5 group
    array : array
    name : string

    """
    if array is not None:
        ds = group.create_dataset(
            name, shape=array.shape, dtype=array.dtype, chunks=True
        )
        ds[:] = array


@deprecate_with_version(
    "dipy.io.peaks.load_peaks is deprecated, Please use"
    "dipy.io.peaks.load_pam instead",
    since="1.10",
    until="1.12",
)
@warning_for_keywords()
def load_peaks(fname, *, verbose=False):
    """Load a PeaksAndMetrics HDF5 file (PAM5).

    Parameters
    ----------
    fname : string
        Filename of PAM5 file.
    verbose : bool
        Print summary information about the loaded file.

    Returns
    -------
    pam : PeaksAndMetrics object

    """
    return load_pam(fname=fname, verbose=verbose)


def load_pam(fname, *, verbose=False):
    """Load a PeaksAndMetrics HDF5 file (PAM5).

    Parameters
    ----------
    fname : string
        Filename of PAM5 file.
    verbose : bool, optional
        Print summary information about the loaded file.

    Returns
    -------
    pam : PeaksAndMetrics object
        Object holding peaks information and metrics.

    """
    if os.path.splitext(fname)[1].lower() != ".pam5":
        raise IOError("This function supports only PAM5 (HDF5) files")

    f = h5py.File(fname, "r")

    pam = PeaksAndMetrics()

    pamh = f["pam"]

    version = f.attrs["version"]

    if version != "0.0.1":
        raise OSError(f"Incorrect PAM5 file version {version}")

    peak_dirs = pamh["peak_dirs"][:]
    peak_values = pamh["peak_values"][:]
    peak_indices = pamh["peak_indices"][:]

    sphere_vertices = pamh["sphere_vertices"][:]

    pam.affine = pamh["affine"][:] if "affine" in pamh else None
    pam.peak_dirs = peak_dirs
    pam.peak_values = peak_values
    pam.peak_indices = peak_indices
    pam.shm_coeff = pamh["shm_coeff"][:] if "shm_coeff" in pamh else None
    pam.sphere = Sphere(xyz=sphere_vertices)
    pam.B = pamh["B"][:] if "B" in pamh else None
    pam.total_weight = pamh["total_weight"][:][0] if "total_weight" in pamh else None
    pam.ang_thr = pamh["ang_thr"][:][0] if "ang_thr" in pamh else None
    pam.gfa = pamh["gfa"][:] if "gfa" in pamh else None
    pam.qa = pamh["qa"][:] if "qa" in pamh else None
    pam.odf = pamh["odf"][:] if "odf" in pamh else None

    f.close()

    if verbose:
        print("PAM5 version")
        print(version)
        print("Affine")
        print(pam.affine)
        print("Dirs shape")
        print(pam.peak_dirs.shape)
        print("SH shape")
        if pam.shm_coeff is not None:
            print(pam.shm_coeff.shape)
        else:
            print("None")
        print("ODF shape")
        if pam.odf is not None:
            print(pam.odf.shape)
        else:
            print("None")
        print("Total weight")
        print(pam.total_weight)
        print("Angular threshold")
        print(pam.ang_thr)
        print("Sphere vertices shape")
        print(pam.sphere.vertices.shape)

    return pam


@deprecate_with_version(
    "dipy.io.peaks.save_peaks is deprecated, Please use "
    "dipy.io.peaks.save_pam instead",
    since="1.10.0",
    until="1.12.0",
)
@warning_for_keywords()
def save_peaks(fname, pam, *, affine=None, verbose=False):
    """Save PeaksAndMetrics object attributes in a PAM5 file (HDF5).

    Parameters
    ----------
    fname : string
        Filename of PAM5 file.
    pam : PeaksAndMetrics
        Object holding peak_dirs, shm_coeffs and other attributes.
    affine : array
        The 4x4 matrix transforming the date from native to world coordinates.
        PeaksAndMetrics should have that attribute but if not it can be
        provided here. Default None.
    verbose : bool
        Print summary information about the saved file.

    """
    return save_pam(fname=fname, pam=pam, affine=affine, verbose=verbose)


def save_pam(fname, pam, *, affine=None, verbose=False):
    """Save all important attributes of object PeaksAndMetrics in a PAM5 file (HDF5).

    Parameters
    ----------
    fname : str
        Filename of PAM5 file.
    pam : PeaksAndMetrics
        Object holding peaks information and metrics.
    affine : ndarray, optional
        The 4x4 matrix transforming the date from native to world coordinates.
        PeaksAndMetrics should have that attribute but if not it can be
        provided here.
    verbose : bool, optional
        Print summary information about the saved file.

    """
    if os.path.splitext(fname)[1] != ".pam5":
        raise IOError("This function saves only PAM5 (HDF5) files")

    if not (
        hasattr(pam, "peak_dirs")
        and hasattr(pam, "peak_values")
        and hasattr(pam, "peak_indices")
    ):
        msg = "Cannot save object without peak_dirs, peak_values"
        msg += " and peak_indices"
        raise ValueError(msg)

    if not (
        isinstance(pam.peak_dirs, np.ndarray)
        and isinstance(pam.peak_values, np.ndarray)
        and isinstance(pam.peak_indices, np.ndarray)
    ):
        msg = "Cannot save object: peak_dirs, peak_values"
        msg += " and peak_indices should be a ndarray"
        raise ValueError(msg)

    f = h5py.File(fname, "w")

    group = f.create_group("pam")
    f.attrs["version"] = "0.0.1"

    version_string = f.attrs["version"]

    affine = pam.affine if hasattr(pam, "affine") else affine
    shm_coeff = pam.shm_coeff if hasattr(pam, "shm_coeff") else None
    odf = pam.odf if hasattr(pam, "odf") else None
    vertices = None
    if hasattr(pam, "sphere") and pam.sphere is not None:
        vertices = pam.sphere.vertices

    _safe_save(group, affine, "affine")
    _safe_save(group, pam.peak_dirs, "peak_dirs")
    _safe_save(group, pam.peak_values, "peak_values")
    _safe_save(group, pam.peak_indices, "peak_indices")
    _safe_save(group, shm_coeff, "shm_coeff")
    _safe_save(group, vertices, "sphere_vertices")
    _safe_save(group, pam.B, "B")
    _safe_save(group, np.array([pam.total_weight]), "total_weight")
    _safe_save(group, np.array([pam.ang_thr]), "ang_thr")
    _safe_save(group, pam.gfa, "gfa")
    _safe_save(group, pam.qa, "qa")
    _safe_save(group, odf, "odf")

    f.close()

    if verbose:
        print("PAM5 version")
        print(version_string)
        print("Affine")
        print(affine)
        print("Dirs shape")
        print(pam.peak_dirs.shape)
        print("SH shape")
        if shm_coeff is not None:
            print(shm_coeff.shape)
        else:
            print("None")
        print("ODF shape")
        if odf is not None:
            print(pam.odf.shape)
        else:
            print("None")
        print("Total weight")
        print(pam.total_weight)
        print("Angular threshold")
        print(pam.ang_thr)
        print("Sphere vertices shape")
        print(pam.sphere.vertices.shape)

    return pam


@deprecate_with_version(
    "dipy.io.peaks.peaks_to_niftis is deprecated, Please"
    " use dipy.io.peaks.pam_to_niftis instead",
    since="1.10.0",
    until="1.12.0",
)
@warning_for_keywords()
def peaks_to_niftis(
    pam,
    fname_shm,
    fname_dirs,
    fname_values,
    fname_indices,
    *,
    fname_gfa=None,
    reshape_dirs=False,
):
    """Save SH, directions, indices and values of peaks to Nifti.

    Parameters
    ----------
    pam : PeaksAndMetrics
        Object holding peaks information and metrics.
    fname_shm : str
        Spherical Harmonics coefficients filename.
    fname_dirs : str
        Peaks direction filename.
    fname_values : str
        Peaks values filename.
    fname_indices : str
        Peaks indices filename.
    fname_gfa : str, optional
        Generalized FA filename.
    reshape_dirs : bool, optional
        If True, reshape peaks for visualization.

    """
    return pam_to_niftis(
        pam=pam,
        fname_peaks_dir=fname_dirs,
        fname_peaks_values=fname_values,
        fname_peaks_indices=fname_indices,
        fname_gfa=fname_gfa,
        fname_shm=fname_shm,
        reshape_dirs=reshape_dirs,
    )


def pam_to_niftis(
    pam,
    *,
    fname_peaks_dir="peaks_dirs.nii.gz",
    fname_peaks_values="peaks_values.nii.gz",
    fname_peaks_indices="peaks_indices.nii.gz",
    fname_shm="shm.nii.gz",
    fname_gfa="gfa.nii.gz",
    fname_sphere="sphere.txt",
    fname_b="B.nii.gz",
    fname_qa="qa.nii.gz",
    reshape_dirs=False,
):
    """Save SH, directions, indices and values of peaks to Nifti.

    Parameters
    ----------
    pam : PeaksAndMetrics
        Object holding peaks information and metrics.
    fname_peaks_dir : str, optional
        Peaks direction filename.
    fname_peaks_values : str, optional
        Peaks values filename.
    fname_peaks_indices : str, optional
        Peaks indices filename.
    fname_shm : str, optional
        Spherical Harmonics coefficients filename. It will be saved if available.
    fname_gfa : str, optional
        Generalized FA filename. It will be saved if available.
    fname_sphere : str, optional
        Sphere vertices filename. It will be saved if available.
    fname_b : str, optional
        B Matrix filename. Matrix that transforms spherical harmonics to
        spherical function. It will be saved if available.
    fname_qa : str, optional
        Quantitative Anisotropy filename. It will be saved if available.
    reshape_dirs : bool, optional
        If True, Reshape and convert to float32 a set of peaks for
        visualisation with mrtrix or the fibernavigator.

    """
    if reshape_dirs:
        pam_dirs = reshape_peaks_for_visualization(pam)
    else:
        pam_dirs = pam.peak_dirs.astype(np.float32)

    save_nifti(fname_peaks_dir, pam_dirs, pam.affine)
    save_nifti(fname_peaks_values, pam.peak_values.astype(np.float32), pam.affine)
    save_nifti(fname_peaks_indices, pam.peak_indices, pam.affine)

    for attr, fname in [("gfa", fname_gfa), ("B", fname_b), ("qa", fname_qa)]:
        obj = getattr(pam, attr, None)
        if obj is None:
            continue
        save_nifti(fname, obj, pam.affine)

    if hasattr(pam, "shm_coeff") and pam.shm_coeff is not None:
        save_nifti(fname_shm, pam.shm_coeff.astype(np.float32), pam.affine)
    if hasattr(pam, "sphere") and pam.sphere is not None:
        np.savetxt(fname_sphere, pam.sphere.vertices)


def niftis_to_pam(
    affine,
    peak_dirs,
    peak_values,
    peak_indices,
    *,
    shm_coeff=None,
    sphere=None,
    gfa=None,
    B=None,
    qa=None,
    odf=None,
    total_weight=None,
    ang_thr=None,
    pam_file=None,
):
    """Return SH, directions, indices and values of peaks to pam5.

    Parameters
    ----------
    affine : array, (4, 4)
        The matrix defining the affine transform.
    peak_dirs : ndarray
        The direction of each peak.
    peak_values : ndarray
        The value of the peaks.
    peak_indices : ndarray
        Indices (in sphere vertices) of the peaks in each voxel.
    shm_coeff : array, optional
        Spherical harmonics coefficients.
    sphere : `Sphere` class instance, optional
        The sphere providing discrete directions for evaluation.
    gfa : ndarray, optional
        Generalized FA volume.
    B : ndarray, optional
        Matrix that transforms spherical harmonics to spherical function.
    qa : array, optional
        Quantitative Anisotropy in each voxel.
    odf : ndarray, optional
        SH coefficients for the ODF spherical function.
    total_weight : float, optional
        Total weight of the peaks.
    ang_thr : float, optional
        Angular threshold of the peaks.
    pam_file : str, optional
        Filename of the desired pam file.

    Returns
    -------
    pam : PeaksAndMetrics
        Object holding peak_dirs, shm_coeffs and other attributes.

    """
    pam = PeaksAndMetrics()
    pam.affine = affine
    pam.peak_dirs = peak_dirs
    pam.peak_values = peak_values
    pam.peak_indices = peak_indices

    for name, value in [
        ("shm_coeff", shm_coeff),
        ("sphere", sphere),
        ("B", B),
        ("total_weight", total_weight),
        ("ang_thr", ang_thr),
        ("gfa", gfa),
        ("qa", qa),
        ("odf", odf),
    ]:
        if value is not None:
            setattr(pam, name, value)

    if pam_file:
        save_pam(pam_file, pam)
    return pam


def tensor_to_pam(
    evals,
    evecs,
    affine,
    *,
    shm_coeff=None,
    sphere=None,
    gfa=None,
    B=None,
    qa=None,
    odf=None,
    total_weight=None,
    ang_thr=None,
    pam_file=None,
    npeaks=5,
    generate_peaks_indices=True,
):
    """Convert diffusion tensor to pam5.

    Parameters
    ----------
    evals : ndarray
        Eigenvalues of a diffusion tensor. shape should be (...,3).
    evecs : ndarray
        Eigen vectors from the tensor model.
    affine : array, (4, 4)
        The matrix defining the affine transform.
    shm_coeff : array, optional
        Spherical harmonics coefficients.
    sphere : `Sphere` class instance, optional
        The sphere providing discrete directions for evaluation.
    gfa : ndarray, optional
        Generalized FA volume.
    B : ndarray, optional
        Matrix that transforms spherical harmonics to spherical function.
    qa : array, optional
        Quantitative Anisotropy in each voxel.
    odf : ndarray, optional
        SH coefficients for the ODF spherical function.
    pam_file : str, optional
        Filename of the desired pam file.
    npeaks : int, optional
        Maximum number of peaks found.
    generate_peaks_indices : bool, optional
    total_weight : float, optional
        Total weight of the peaks.
    ang_thr : float, optional
        Angular threshold of the peaks.

    Returns
    -------
    pam : PeaksAndMetrics
        Object holding peaks information and metrics.
    """
    npeaks = 1 if npeaks < 1 else npeaks
    npeaks = min(npeaks, evals.shape[-1])
    shape = evals.shape[:3]
    peaks_dirs = np.zeros((shape + (npeaks, 3)))
    peaks_dirs[..., :npeaks, :] = evecs[..., :npeaks, :]
    peaks_values = np.zeros((shape + (npeaks,)))
    peaks_values[..., :npeaks] = evals[..., :npeaks]

    if generate_peaks_indices:
        vertices = sphere.vertices if sphere else None
        peaks_indices = quantize_evecs(evecs[..., :npeaks, :], odf_vertices=vertices)
    else:
        peaks_indices = np.zeros((shape + (npeaks,)), dtype="int")
        peaks_indices.fill(-1)

    return niftis_to_pam(
        affine=affine,
        peak_dirs=peaks_dirs,
        peak_values=peaks_values,
        peak_indices=peaks_indices.astype(np.int32),
        shm_coeff=shm_coeff,
        sphere=sphere,
        gfa=gfa,
        B=B,
        qa=qa,
        odf=odf,
        total_weight=total_weight,
        ang_thr=ang_thr,
        pam_file=pam_file,
    )



import warnings

import numpy as np
from scipy.sparse import coo_array
from scipy.spatial import cKDTree

from dipy.io.stateful_tractogram import StatefulTractogram
from dipy.segment.metric import mean_euclidean_distance
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import set_number_of_points


class FastStreamlineSearch:
    @warning_for_keywords()
    def __init__(
        self,
        ref_streamlines,
        max_radius,
        *,
        nb_mpts=4,
        bin_size=20.0,
        resampling=24,
        bidirectional=True,
    ):
        """Fast Streamline Search (FFS)

        Generate the Binned K-D Tree structure with reference streamlines,
        using streamlines barycenter and mean-points.
        See :footcite:p:`StOnge2022` for further details.

        Parameters
        ----------
        ref_streamlines : Streamlines
            Streamlines (ref) to generate the tree structure.
        max_radius : float
            The maximum radius (distance) for subsequent streamline search.
            Used to compute the overlap in-between bins.
        nb_mpts : int, optional
            Number of means points to improve computation speed.
            (this only changes computation time)
        bin_size : float, optional
            The bin size to separate streamlines in groups.
            (this only changes computation time)
        resampling : int, optional
            Number of points used to reshape each streamline.
        bidirectional : bool, optional
            Compute the smallest distance with and without flip.

        Notes
        -----
        Make sure that streamlines are aligned in the same space.
        Preferably in millimeter space (voxmm or rasmm).

        References
        ----------
        .. footbibliography::
        """
        if max_radius <= 0.0:
            raise ValueError("max_radius needs to be a positive value")

        if resampling < 20:
            warnings.warn(
                "For accurate results, resampling should be"
                " at least >= 10 and preferably >= 20",
                stacklevel=2,
            )

        if resampling % nb_mpts != 0:
            raise ValueError("nb_mpts needs to be a factor of resampling")

        if isinstance(ref_streamlines, StatefulTractogram):
            ref_streamlines = ref_streamlines.streamlines

        self.nb_mpts = nb_mpts
        self.bin_size = bin_size
        self.bidirectional = bidirectional
        self.resampling = resampling
        self.max_radius = max_radius

        # Resample streamlines
        self.ref_slines = self._resample(ref_streamlines)
        self.ref_nb_slines = len(self.ref_slines)

        if self.bidirectional:
            self.ref_slines = np.concatenate(
                [self.ref_slines, np.flip(self.ref_slines, axis=1)]
            )

        # Compute streamlines barycenter
        barycenters = self._slines_barycenters(self.ref_slines)

        # Compute bin shape (min, max, shape)
        bin_overlap = max_radius
        self.min_box = np.min(barycenters, axis=0) - bin_overlap
        self.max_box = np.max(barycenters, axis=0) + bin_overlap

        box_length = self.max_box - self.min_box
        self.bin_shape = (box_length // bin_size).astype(int) + 1

        # Compute the center of each bin
        bin_list = np.arange(np.prod(self.bin_shape))
        all_bins = np.vstack(np.unravel_index(bin_list, self.bin_shape)).T
        bins_center = all_bins * bin_size + self.min_box + bin_size / 2.0

        # Assign a list of streamlines to each bin
        baryc_tree = cKDTree(barycenters)
        center_dist = bin_size / 2.0 + bin_overlap
        baryc_bins = baryc_tree.query_ball_point(bins_center, center_dist, p=np.inf)

        # Compute streamlines mean-points
        meanpts = self._slines_mean_points(self.ref_slines)

        # Compute bin indices, streamlines + mean-points tree
        self.bin_dict = {}
        for i, baryc_b in enumerate(baryc_bins):
            if baryc_b:
                slines_id = np.asarray(baryc_b)
                self.bin_dict[i] = (slines_id, cKDTree(meanpts[slines_id]))

    @warning_for_keywords()
    def radius_search(self, streamlines, radius, *, use_negative=True):
        """Radius Search using Fast Streamline Search

        For each given streamlines, return all reference streamlines
        within the given radius. See :footcite:p:`StOnge2022` for further
        details.

        Parameters
        ----------
        streamlines : Streamlines
            Streamlines to generate the tree structure.
        radius : float
            Search radius (with MDF / average L2 distance)
            must be smaller than max_radius when FFS was initialized.
        use_negative : bool, optional
            When used with bidirectional,
            negative values are returned for reversed order neighbors.


        Returns
        -------
        res : scipy COOrdinates sparse matrix (nb_slines x nb_slines_ref)
            Adjacency matrix containing all neighbors within the given radius

        Notes
        -----
        Given streamlines should be already aligned with ref streamlines.
        Preferably in millimeter space (voxmm or rasmm).

        References
        ----------
        .. footbibliography::
        """
        if radius > self.max_radius:
            raise ValueError(
                "radius should be smaller or equal to the given"
                "\n 'max_radius' in FastStreamlineSearch init"
            )

        if isinstance(streamlines, StatefulTractogram):
            streamlines = streamlines.streamlines

        # Resample query streamlines
        q_slines = self._resample(streamlines)
        q_nb_slines = len(q_slines)

        # Compute streamlines barycenter
        q_baryc = self._slines_barycenters(q_slines)

        # Verify if each barycenter are inside the min max box
        u_bin, binned_slines_ids = self._barycenters_binning(q_baryc)

        # Adapting radius for L1 query: sqrt(3) = 1.73205080756887729..
        # Rounded up for float32 precision to avoid error / false negative
        l1_sum_dist = 1.73205081 * radius * self.nb_mpts

        # Search for all similar streamlines
        list_id = []
        list_id_ref = []
        list_dist = []
        for i, bin_id in enumerate(u_bin):
            if bin_id in self.bin_dict:
                slines_id_ref, ref_tree = self.bin_dict[bin_id]
                slines_id = binned_slines_ids[i]

                mpts = self._slines_mean_points(q_slines[slines_id])

                # Compute Tree L1 Query with mean-points
                res = ref_tree.query_ball_point(mpts, l1_sum_dist, p=1)

                # Refine distance with the complete
                for s, ref_ids in enumerate(res):
                    if ref_ids:
                        s_id = slines_id[s]
                        rs_ids = slines_id_ref[ref_ids]
                        d = mean_euclidean_distance(
                            q_slines[s_id], self.ref_slines[rs_ids]
                        )

                        # Return all pairs within the radius
                        in_dist_max = d < radius
                        id_ref = rs_ids[in_dist_max]
                        id_s = np.full_like(id_ref, s_id)

                        list_id.append(id_s)
                        list_id_ref.append(id_ref)
                        list_dist.append(d[in_dist_max])

        # Combine all results in a coup sparse matrix
        if len(list_id) > 0:
            ids_in = np.hstack(list_id)
            ids_ref = np.hstack(list_id_ref)
            dist = np.hstack(list_dist)

            if self.bidirectional:
                flipped = ids_ref >= self.ref_nb_slines
                ids_ref[flipped] -= self.ref_nb_slines
                if use_negative:
                    dist[flipped] *= -1.0

            return coo_array(
                (dist, (ids_in, ids_ref)), shape=(q_nb_slines, self.ref_nb_slines)
            )

        # No results, return an empty sparse matrix
        return coo_array((q_nb_slines, self.ref_nb_slines))

    def _resample(self, streamlines):
        """Resample streamlines"""
        s = np.zeros([len(streamlines), self.resampling, 3], dtype=np.float32)
        for i, sline in enumerate(streamlines):
            if len(sline) < 2:
                s[i] = sline
            else:
                s[i] = set_number_of_points(sline, nb_points=self.resampling)
        return s

    def _slines_barycenters(self, slines_arr):
        """Compute streamlines barycenter"""
        return np.mean(slines_arr, axis=1)

    def _slines_mean_points(self, slines_arr):
        """Compute streamlines mean-points"""
        r_arr = slines_arr.reshape((len(slines_arr), self.nb_mpts, -1, 3))
        mpts = np.mean(r_arr, axis=2)
        return mpts.reshape(len(slines_arr), -1)

    def _barycenters_binning(self, barycenters):
        """Bin indices in a list according to their barycenter position"""
        in_bin = np.logical_and(
            np.all(barycenters >= self.min_box, axis=1),
            np.all(barycenters <= self.max_box, axis=1),
        )

        baryc_to_box = barycenters[in_bin] - self.min_box
        baryc_bins_id = (baryc_to_box // self.bin_size).astype(int)
        baryc_multiid = np.ravel_multi_index(baryc_bins_id.T, self.bin_shape)

        sort_id = np.argsort(baryc_multiid)
        u_bin, mapping = np.unique(baryc_multiid[sort_id], return_index=True)
        slines_ids = np.split(np.flatnonzero(in_bin)[sort_id], mapping[1:])
        return u_bin, slines_ids


def nearest_from_matrix_row(coo_array):
    """
    Return the nearest (smallest) for each row given an coup sparse matrix

    Parameters
    ----------
    coo_array : scipy COOrdinates sparse array (nb_slines x nb_slines_ref)
        Adjacency matrix containing all neighbors within the given radius

    Returns
    -------
    non_zero_ids : numpy array (nb_non_empty_row x 1)
        Indices of each non-empty slines (row)
    nearest_id : numpy array (nb_non_empty_row x 1)
        Indices of the nearest reference match (column)
    nearest_dist : numpy array (nb_non_empty_row x 1)
        Distance for each nearest match
    """
    non_zero_ids = np.unique(coo_array.row)
    sparse_matrix = np.abs(coo_array.tocsr())
    upper_limit = np.max(sparse_matrix.data) + 1.0
    sparse_matrix.data = upper_limit - sparse_matrix.data
    nearest_id = np.squeeze(sparse_matrix.argmax(axis=1).data)[non_zero_ids]
    nearest_dist = upper_limit - np.squeeze(sparse_matrix.max(axis=1).data)
    return non_zero_ids, nearest_id, nearest_dist


def nearest_from_matrix_col(coo_array):
    """
    Return the nearest (smallest) for each column given an coup sparse matrix

    Parameters
    ----------
    coo_array : scipy COOrdinates sparse matrix (nb_slines x nb_slines_ref)
        Adjacency matrix containing all neighbors within the given radius

    Returns
    -------
    non_zero_ids : numpy array (nb_non_empty_col x 1)
        Indices of each non-empty reference (column)
    nearest_id : numpy array (nb_non_empty_col x 1)
        Indices of the nearest slines match (row)
    nearest_dist : numpy array (nb_non_empty_col x 1)
        Distance for each nearest match
    """
    non_zero_ids = np.unique(coo_array.col)
    sparse_matrix = np.abs(coo_array.tocsc())
    upper_limit = np.max(sparse_matrix.data) + 1.0
    sparse_matrix.data = upper_limit - sparse_matrix.data
    nearest_id = np.squeeze(sparse_matrix.argmax(axis=0).data)[non_zero_ids]
    nearest_dist = upper_limit - np.squeeze(sparse_matrix.max(axis=0).data)
    return non_zero_ids, nearest_id, nearest_dist



import os
import tempfile
import warnings

from numpy.testing import assert_equal, assert_raises
import pytest

from dipy.align.streamwarp import bundlewarp, bundlewarp_vector_filed
from dipy.data import read_five_af_bundles, two_cingulum_bundles
from dipy.tracking.streamline import (
    Streamlines,
    set_number_of_points,
    unlist_streamlines,
)
from dipy.utils.optpkg import optional_package

_, have_matplotlib, _ = optional_package("matplotlib")
fury, have_fury, _ = optional_package("fury", min_version="0.10.0")
pd, have_pd, _ = optional_package("pandas")

if have_fury:
    from dipy.viz import window
    from dipy.viz.streamline import (
        sagittal_deprecation_warning_msg,
        show_bundles,
        viz_displacement_mag,
        viz_two_bundles,
        viz_vector_field,
    )

bundles = read_five_af_bundles()


@pytest.mark.skipif(
    not have_fury or not have_matplotlib, reason="Requires FURY and Matplotlib"
)
def test_output_created():
    colors = [
        [0.91, 0.26, 0.35],
        [0.99, 0.50, 0.38],
        [0.99, 0.88, 0.57],
        [0.69, 0.85, 0.64],
        [0.51, 0.51, 0.63],
    ]

    with tempfile.TemporaryDirectory() as temp_dir:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=sagittal_deprecation_warning_msg,
                category=DeprecationWarning,
            )

            view = "sagital"  # codespell:ignore sagital
            fname = os.path.join(temp_dir, f"test_{view}.png")
            show_bundles(bundles, interactive=False, view=view, save_as=fname)
            assert_equal(os.path.exists(fname), True)

        views = ["axial", "sagittal", "coronal"]

        for view in views:
            fname = os.path.join(temp_dir, f"test_{view}.png")
            show_bundles(bundles, interactive=False, view=view, save_as=fname)
            assert_equal(os.path.exists(fname), True)

        fname = os.path.join(temp_dir, "test_colors.png")
        show_bundles(bundles, interactive=False, colors=colors, save_as=fname)
        assert_equal(os.path.exists(fname), True)

        # Check rendered image is not empty
        report = window.analyze_snapshot(fname, find_objects=True)
        assert_equal(report.objects > 0, True)

        cb1, cb2 = two_cingulum_bundles()

        fname = os.path.join(temp_dir, "test_two_bundles.png")
        viz_two_bundles(cb1, cb2, fname=fname)
        assert_equal(os.path.exists(fname), True)


@pytest.mark.skipif(not have_fury, reason="Requires FURY")
def test_incorrect_view():
    assert_raises(
        ValueError, show_bundles, bundles, interactive=False, view="wrong_view"
    )


@pytest.mark.skipif(
    not have_fury or not have_matplotlib or not have_pd,
    reason="Requires FURY, Matplotlib and Pandas",
)
def test_bundlewarp_viz():
    with tempfile.TemporaryDirectory() as temp_dir:
        cingulum_bundles = two_cingulum_bundles()

        cb1 = cingulum_bundles[0]
        cb1 = Streamlines(set_number_of_points(cb1, nb_points=20))

        cb2 = cingulum_bundles[1]
        cb2 = Streamlines(set_number_of_points(cb2, nb_points=20))

        deformed_bundle, affine_bundle, _, _, _ = bundlewarp(cb1, cb2)

        offsets, directions, colors = bundlewarp_vector_filed(
            affine_bundle, deformed_bundle
        )
        points_aligned, _ = unlist_streamlines(affine_bundle)

        fname = os.path.join(temp_dir, "test_vector_field.png")
        viz_vector_field(points_aligned, directions, colors, offsets, fname)
        assert_equal(os.path.exists(fname), True)

        fname = os.path.join(temp_dir, "test_mag_viz.png")
        viz_displacement_mag(affine_bundle, offsets, fname)
        assert_equal(os.path.exists(fname), True)



"""Utilities for calculating voxel adjacencies and neighbourhoods"""

import numpy as np
from scipy.spatial.distance import pdist, squareform

from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def adjacency_calc(img_shape, *, mask=None, cutoff=1.99):
    """Create adjacency list for voxels, accounting for the mask.

    Parameters
    ----------

    img_shape : list
        Spatial shape of the image data.

    mask : array, optional
        A boolean array used to mark the coordinates in the data that
        should be analyzed that should have the same shape as the images.

    cutoff : float, optional
        Cutoff distance in voxel coordinates for finding adjacent voxels.
        e.g. cutoff=2 gives a 3x3 patch, cutoff=2.1 gives a 3x3 patch
        with additional 4 voxels above, below, left, right of this patch.

    Returns
    -------

    adj : list
        List, one entry per voxel, giving the indices of adjacent voxels.
        The list will correspond to the data array once masked and flattened.

    """

    # Image-space coordinates of voxels, flattened
    XYZ = np.meshgrid(*[range(ds) for ds in img_shape], indexing="ij")
    XYZ = np.column_stack([xyz.ravel() for xyz in XYZ])
    dists = squareform(pdist(XYZ))
    dists = dists < cutoff  # adjacency list contains current voxel
    adj = []
    if mask is not None:
        flat_mask = mask.reshape(-1)
        for idx in range(dists.shape[0]):
            if flat_mask[idx]:
                cond = dists[idx, :]
                cond = cond * flat_mask
                cond = cond[flat_mask == 1]  # so indices will match masked array
                adj.append(np.argwhere(cond).flatten().tolist())
    else:
        for idx in range(dists.shape[0]):
            cond = dists[idx, :]
            adj.append(np.argwhere(cond).flatten().tolist())

    return adj



cimport cython
from cython.view cimport array as cvarray
from libc.math cimport sqrt, exp
import numpy as np
cimport numpy as cnp

__all__ = ['firdn', 'upfir', 'nlmeans_block']

cdef inline int _int_max(int a, int b):
    return a if a >= b else b
cdef inline int _int_min(int a, int b):
    return a if a <= b else b


def _firdn_vector(double[:] f, double[:] h, double[:] out):
    cdef int n = len(f)
    cdef int klen = len(h)
    cdef int outLen = (n + klen) // 2
    cdef double ss
    cdef int i, k, limInf, limSup, x = 0, ox = 0, ks = 0
    for i in range(outLen):
        ss = 0
        limInf = _int_max(0, x - klen + 1)
        limSup = 1 + _int_min(n - 1, x)
        ks = limInf
        for k in range(limInf, limSup):
            ss += f[ks] * h[x - k]
            ks += 1
        out[ox] = ss
        x += 2
        ox += 1


def _upfir_vector(double[:] f, double[:] h, double[:] out):
    cdef int n = f.shape[0]
    cdef int klen = h.shape[0]
    cdef int outLen = 2 * n + klen - 2
    cdef int x, limInf, limSup, k, ks
    cdef double ss
    for x in range(outLen):
        limInf = _int_max(0, x - klen + 1)
        if limInf % 2 == 1:
            limInf += 1
        limSup = _int_min(2 * (n - 1), x)
        if limSup % 2 == 1:
            limSup -= 1
        ss = 0
        k = limInf
        ks = limInf // 2
        while k <= limSup:
            ss += f[ks] * h[x - k]
            k += 2
            ks += 1
        out[x] = ss


def _firdn_matrix(double[:, :] F, double[:] h, double[:, :] out):
    cdef int n = F.shape[0]
    cdef int m = F.shape[1]
    cdef int j
    for j in range(m):
        _firdn_vector(F[:, j], h, out[:, j])


def _upfir_matrix(double[:, :] F, double[:] h, double[:, :] out):
    cdef int n = F.shape[0]
    cdef int m = F.shape[1]
    for j in range(m):
        _upfir_vector(F[:, j], h, out[:, j])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void _average_block(double[:, :, :] ima, int x, int y, int z,
                         double[:, :, :] average, double weight) noexcept nogil:
    """
    Computes the weighted average of the patches in a blockwise manner

    Parameters
    ----------
    ima : 3D array of doubles
        input image
    x : integer
        x coordinate of the center voxel
    y : integer
        y coordinate of the center voxel
    z : integer
        z coordinate of the center voxel
    average : 3D array of doubles
        the image where averages are stored
    weight : double
        weight for the weighted averaging
    """

    cdef int a, b, c, x_pos, y_pos, z_pos
    cdef int is_outside
    cdef int neighborhoodsize = average.shape[0] // 2
    for a in range(average.shape[0]):
        for b in range(average.shape[1]):
            for c in range(average.shape[2]):
                x_pos = x + a - neighborhoodsize
                y_pos = y + b - neighborhoodsize
                z_pos = z + c - neighborhoodsize
                is_outside = 0
                if x_pos < 0 or x_pos >= ima.shape[1]:
                    is_outside = 1
                if y_pos < 0 or y_pos >= ima.shape[0]:
                    is_outside = 1
                if z_pos < 0 or z_pos >= ima.shape[2]:
                    is_outside = 1
                if is_outside == 1:
                    average[a, b, c] += weight * (ima[y, x, z]**2)
                else:
                    average[a, b, c] += weight * (ima[y_pos, x_pos, z_pos]**2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void _value_block(double[:, :, :] estimate, double[:, :, :] Label, int x, int y,
                       int z, double[:, :, :] average, double global_sum, double hh, int rician_int) noexcept nogil:

    """
    Computes the final estimate of the denoised image

    Parameters
    ----------
    estimate : 3D array of doubles
        The denoised estimate array
    Label : 3D array of doubles
        The label map for block wise weighted averaging
    x : integer
        x coordinate of the center voxel
    y : integer
        y coordinate of the center voxel
    z : integer
        z coordinate of the center voxel
    average : 3D array of doubles
        weighted average image
    global_sum : double
        total weight sum
    hh : double
        weight parameter
    rician_int : integer
        0 or 1 as per the boolean value
    """

    cdef int is_outside, a, b, c, x_pos, y_pos, z_pos, count = 0
    cdef double value = 0.0
    cdef double denoised_value = 0.0
    cdef double label = 0.0
    cdef int neighborhoodsize = average.shape[0] // 2
    for a in range(average.shape[0]):
        for b in range(average.shape[1]):
            for c in range(average.shape[2]):
                is_outside = 0
                x_pos = x + a - neighborhoodsize
                y_pos = y + b - neighborhoodsize
                z_pos = z + c - neighborhoodsize
                if x_pos < 0 or x_pos >= estimate.shape[1]:
                    is_outside = 1
                if y_pos < 0 or y_pos >= estimate.shape[0]:
                    is_outside = 1
                if z_pos < 0 or z_pos >= estimate.shape[2]:
                    is_outside = 1
                if is_outside == 0:
                    value = estimate[y_pos, x_pos, z_pos]
                    if rician_int:
                        denoised_value = (average[a, b, c] / global_sum) - hh
                    else:
                        denoised_value = (average[a, b, c] / global_sum)
                    if denoised_value > 0:
                        denoised_value = sqrt(denoised_value)
                    else:
                        denoised_value = 0.0
                    value += denoised_value
                    label = Label[y_pos, x_pos, z_pos]
                    estimate[y_pos, x_pos, z_pos] = value
                    Label[y_pos, x_pos, z_pos] = label + 1


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double _distance(double[:, :, :] image, int x, int y, int z,
                      int nx, int ny, int nz, int block_radius) nogil:
    """
    Computes the distance between two square subpatches of image located at
    p and q, respectively. If the centered squares lie beyond the boundaries
    of image, they are mirrored.

    Parameters
    ----------
    image : 3D array of doubles
        the image whose voxels are taken
    x : integer
        x coordinate of first patch's center
    y : integer
        y coordinate of first patch's center
    z : integer
        z coordinate of first patch's center
    nx : integer
        nx coordinate of second patch's center
    ny : integer
        ny coordinate of second patch's center
    nz : integer
        nz coordinate of second patch's center
    block_radius : integer
        block radius for which the distance is computed for
    """

    cdef double acu, distancetotal
    cdef int i, j, k, ni1, nj1, ni2, nj2, nk1, nk2
    cdef int sx = image.shape[1], sy = image.shape[0], sz = image.shape[2]
    acu = 0
    distancetotal = 0
    for i in range(-block_radius, block_radius + 1):
        for j in range(-block_radius, block_radius + 1):
            for k in range(-block_radius, block_radius + 1):
                ni1 = x + i
                nj1 = y + j
                nk1 = z + k
                ni2 = nx + i
                nj2 = ny + j
                nk2 = nz + k
                if ni1 < 0:
                    ni1 = -ni1
                if nj1 < 0:
                    nj1 = -nj1
                if ni2 < 0:
                    ni2 = -ni2
                if nj2 < 0:
                    nj2 = -nj2
                if nk1 < 0:
                    nk1 = -nk1
                if nk2 < 0:
                    nk2 = -nk2
                if ni1 >= sx:
                    ni1 = 2 * sx - ni1 - 1
                if nj1 >= sy:
                    nj1 = 2 * sy - nj1 - 1
                if nk1 >= sz:
                    nk1 = 2 * sz - nk1 - 1
                if ni2 >= sx:
                    ni2 = 2 * sx - ni2 - 1
                if nj2 >= sy:
                    nj2 = 2 * sy - nj2 - 1
                if nk2 >= sz:
                    nk2 = 2 * sz - nk2 - 1
                distancetotal += (image[nj1, ni1, nk1] -
                                  image[nj2, ni2, nk2])**2
                acu = acu + 1
    return distancetotal / acu


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double _local_mean(double[:, :, :]ima, int x, int y, int z) nogil:
    """
    local mean of a 3x3x3 patch centered at x,y,z
    """
    cdef int dims0 = ima.shape[0]
    cdef int dims1 = ima.shape[1]
    cdef int dims2 = ima.shape[2]
    cdef double ss = 0
    cdef int px, py, pz, dx, dy, dz, nx, ny, nz
    for px in range(x - 1, x + 2):
        for py in range(y - 1, y + 2):
            for pz in range(z - 1, z + 2):
                px = (-px if px < 0 else (2 * dims0 -
                                          px - 1 if px >= dims0 else px))
                py = (-py if py < 0 else (2 * dims1 -
                                          py - 1 if py >= dims1 else py))
                pz = (-pz if pz < 0 else (2 * dims2 -
                                          pz - 1 if pz >= dims2 else pz))
                ss += ima[px, py, pz]

    return ss / 27.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double _local_variance(double[:, :, :] ima, double mean, int x, int y, int z) nogil:
    """
    local variance of a 3x3x3 patch centered at x,y,z
    """
    dims0 = ima.shape[0]
    dims1 = ima.shape[1]
    dims2 = ima.shape[2]
    cdef int cnt = 0
    cdef double ss = 0
    cdef int dx, dy, dz, nx, ny, nz
    for px in range(x - 1, x + 2):
        for py in range(y - 1, y + 2):
            for pz in range(z - 1, z + 2):
                if ((px >= 0 and py >= 0 and pz > 0) and
                        (px < dims0 and py < dims1 and pz < dims2)):
                    ss += (ima[px, py, pz] - mean) * (ima[px, py, pz] - mean)
                    cnt += 1
    return ss / (cnt - 1)

cpdef firdn(double[:, :] image, double[:] h):
    """
    Applies the filter given by the convolution kernel 'h' columnwise to
    'image', then subsamples by 2. This is a special case of the matlab's
    'upfirdn' function, ported to python. Returns the filtered image.

    Parameters
    ----------
    image: 2D array of doubles
        the input image to be filtered
    h: double array
        the convolution kernel
    """
    nrows = image.shape[0]
    ncols = image.shape[1]
    ll = h.shape[0]
    cdef double[:, :] filtered = np.zeros(shape=((nrows + ll) // 2, ncols))
    _firdn_matrix(image, h, filtered)
    return filtered

cpdef upfir(double[:, :] image, double[:] h):
    """
    Upsamples the columns of the input image by 2, then applies the
    convolution kernel 'h' (again, columnwise). This is a special case of the
    matlab's 'upfirdn' function, ported to python. Returns the filtered image.

    Parameters
    ----------
    image: 2D array of doubles
        the input image to be filtered
    h: double array
        the convolution kernel
    """
    nrows = image.shape[0]
    ncols = image.shape[1]
    ll = h.shape[0]
    cdef double[:, :] filtered = np.zeros(shape=(2 * nrows + ll - 2, ncols))
    _upfir_matrix(image, h, filtered)
    return filtered


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def nlmeans_block(double[:, :, :]image, double[:, :, :] mask, int patch_radius, int block_radius, double h, int rician):
    """Non-Local Means Denoising Using Blockwise Averaging.

    See :footcite:p:`Coupe2008` and :footcite:p:`Coupe2012` for further details
    about the method.

    Parameters
    ----------
    image : 3D array of doubles
        the input image, corrupted with rician noise
    mask : 3D array of doubles
        the input mask
    patch_radius :  int
        similar patches in the non-local means are searched for locally,
        inside a cube of side 2*v+1 centered at each voxel of interest.
    block_radius :  int
        the size of the block to be used (2*f+1)x(2*f+1)x(2*f+1) in the
        blockwise non-local means implementation (the Coupe's proposal).
    h :  double
        the estimated amount of rician noise in the input image: in P.
        Coupe et al. the rician noise was simulated as
        sqrt((f+x)^2 + (y)^2) where f is the pixel value and x and y are
        independent realizations of a random variable with Normal
        distribution, with mean=0 and standard deviation=h
    rician : boolean
        If True the noise is estimated as Rician, otherwise Gaussian noise
        is assumed.

    Returns
    -------
    fima: 3D double array
        the denoised output which has the same shape as input image.

    References
    ----------
    .. footbibliography::

    """

    cdef int[:] dims = cvarray((3,), itemsize=sizeof(int), format="i")
    dims[0] = image.shape[0]
    dims[1] = image.shape[1]
    dims[2] = image.shape[2]
    cdef double hh = 2 * h * h
    cdef int Ndims = (2 * block_radius + 1)**3
    cdef int nvox = dims[0] * dims[1] * dims[2]
    cdef double[:, :, :] average = np.zeros((2 * block_radius + 1, 2 * block_radius + 1, 2 * block_radius + 1), dtype=np.float64)
    cdef double[:, :, :] fima = np.zeros_like(image)
    cdef double[:, :, :] means = np.zeros_like(image)
    cdef double[:, :, :] variances = np.zeros_like(image)
    cdef double[:, :, :] Estimate = np.zeros_like(image)
    cdef double[:, :, :] Label = np.zeros_like(image)
    cdef cnp.npy_intp i, j, k, ni, nj, nk
    cdef double t1, t2
    cdef double epsilon = 0.00001
    cdef double mu1 = 0.95
    cdef double var1 = 0.5 + 1e-7
    cdef double d
    cdef double totalWeight, wmax, w

    with nogil:
        for k in range(dims[2]):
            for i in range(dims[1]):
                for j in range(dims[0]):
                    means[j, i, k] = _local_mean(image, j, i, k)
                    variances[j, i, k] = _local_variance(
                        image, means[j, i, k], j, i, k)
        for k in range(0, dims[2], 2):
            for i in range(0, dims[1], 2):
                for j in range(0, dims[0], 2):
                    with gil:
                        average[...] = 0
                    totalWeight = 0
                    if (means[j, i, k] <= epsilon) or (
                            variances[j, i, k] <= epsilon):
                        wmax = 1.0
                        _average_block(image, i, j, k, average, wmax)
                        totalWeight += wmax
                        _value_block(Estimate, Label, i, j, k,
                                     average, totalWeight, hh, rician)
                    else:
                        wmax = 0
                        for nk in range(k - patch_radius, k + patch_radius + 1):
                            for ni in range(i - patch_radius, i + patch_radius + 1):
                                for nj in range(j - patch_radius, j + patch_radius + 1):
                                    if ni == i and nj == j and nk == k:
                                        continue
                                    if ni < 0 or nj < 0 or nk < 0 or nj >= dims[0] or ni >= dims[1] or nk >= dims[2]:
                                        continue
                                    if ((means[nj, ni, nk] <= epsilon) or (
                                            variances[nj, ni, nk] <= epsilon)):
                                        continue
                                    t1 = (means[j, i, k]) / (means[nj, ni, nk])
                                    t2 = (variances[j, i, k]) / \
                                        (variances[nj, ni, nk])
                                    if mu1 < t1 < (1 / mu1) and var1 < t2 < (1 / var1):
                                        d = _distance(
                                            image, i, j, k, ni, nj, nk, block_radius)
                                        w = exp(-d / (h * h))
                                        if w > wmax:
                                            wmax = w
                                        _average_block(
                                            image, ni, nj, nk, average, w)
                                        totalWeight += w

                        if totalWeight != 0.0:
                            _value_block(Estimate, Label, i, j, k,
                                         average, totalWeight, hh, rician)

        for k in range(0, dims[2]):
            for i in range(0, dims[1]):
                for j in range(0, dims[0]):

                    if mask[j, i, k] == 0:
                        fima[j, i, k] = 0

                    else:
                        if Label[j, i, k] == 0.0:
                            fima[j, i, k] = image[j, i, k]
                        else:
                            fima[j, i, k] = Estimate[j, i, k] / Label[j, i, k]

    return fima



import numpy as np
import numpy.testing as npt

from dipy.core.sphere import Sphere
from dipy.denoise.enhancement_kernel import EnhancementKernel
from dipy.tracking.fbcmeasures import FBCMeasures


def test_fbc():
    """Test the FBC measures on a set of fibers"""

    # Generate two fibers of 10 points
    streamlines = []
    for i in range(2):
        fiber = np.zeros((10, 3))
        for j in range(10):
            fiber[j, 0] = j
            fiber[j, 1] = i * 0.2
            fiber[j, 2] = 0
            streamlines.append(fiber)

    # Create lookup table.
    # A fixed set of orientations is used to guarantee deterministic results
    D33 = 1.0
    D44 = 0.04
    t = 1
    sphere = Sphere(
        xyz=np.array(
            [
                [0.82819078, 0.51050355, 0.23127074],
                [-0.10761926, -0.95554309, 0.27450957],
                [0.4101745, -0.07154038, 0.90919682],
                [-0.75573448, 0.64854889, 0.09082809],
                [-0.56874549, 0.01377562, 0.8223982],
            ]
        )
    )
    k = EnhancementKernel(D33, D44, t, orientations=sphere, force_recompute=True)

    # run FBC
    fbc = FBCMeasures(streamlines, k, verbose=True)

    # get FBC values
    fbc_sl_orig, clrs_orig, rfbc_orig = fbc.get_points_rfbc_thresholded(
        0, emphasis=0.01
    )

    # check mean RFBC against tested value
    npt.assert_almost_equal(np.mean(rfbc_orig), 1.0500466494329224, decimal=4)



import os.path as op

import nibabel as nib
import numpy as np
import numpy.testing as npt
import scipy.linalg as la

import dipy.core.gradients as grad
import dipy.core.optimize as opt
import dipy.data as dpd
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
import dipy.tracking.life as life

THIS_DIR = op.dirname(__file__)


def test_streamline_gradients():
    streamline = [[1, 2, 3], [4, 5, 6], [5, 6, 7], [8, 9, 10]]
    grads = np.array([[3, 3, 3], [2, 2, 2], [2, 2, 2], [3, 3, 3]])
    npt.assert_array_equal(life.streamline_gradients(streamline), grads)


def test_streamline_tensors():
    # Small streamline
    streamline = [[1, 2, 3], [4, 5, 3], [5, 6, 3]]
    # Non-default eigenvalues:
    evals = [0.0012, 0.0006, 0.0004]
    streamline_tensors = life.streamline_tensors(streamline, evals=evals)
    npt.assert_array_almost_equal(
        streamline_tensors[0],
        np.array([[0.0009, 0.0003, 0.0], [0.0003, 0.0009, 0.0], [0.0, 0.0, 0.0004]]),
    )

    # Get the eigenvalues/eigenvectors:
    eigvals, eigvecs = la.eig(streamline_tensors[0])
    eigvecs = eigvecs[np.argsort(eigvals)[::-1]]
    eigvals = eigvals[np.argsort(eigvals)[::-1]]

    npt.assert_array_almost_equal(eigvals, np.array([0.0012, 0.0006, 0.0004]))

    npt.assert_array_almost_equal(eigvecs[0], np.array([0.70710678, -0.70710678, 0.0]))
    # Another small streamline
    streamline = [[1, 0, 0], [2, 0, 0], [3, 0, 0]]
    streamline_tensors = life.streamline_tensors(streamline, evals=evals)

    for t in streamline_tensors:
        eigvals, eigvecs = la.eig(t)
        eigvecs = eigvecs[np.argsort(eigvals)[::-1]]
        # This one has no rotations - all tensors are simply the canonical:
        npt.assert_almost_equal(np.rad2deg(np.arccos(np.dot(eigvecs[0], [1, 0, 0]))), 0)
        npt.assert_almost_equal(np.rad2deg(np.arccos(np.dot(eigvecs[1], [0, 1, 0]))), 0)
        npt.assert_almost_equal(np.rad2deg(np.arccos(np.dot(eigvecs[2], [0, 0, 1]))), 0)


def test_streamline_signal():
    data_file, bval_file, bvec_file = dpd.get_fnames(name="small_64D")
    gtab = grad.gradient_table(bval_file, bvecs=bvec_file)
    evals = [0.0015, 0.0005, 0.0005]
    streamline1 = [
        [[1, 2, 3], [4, 5, 3], [5, 6, 3], [6, 7, 3]],
        [[1, 2, 3], [4, 5, 3], [5, 6, 3]],
    ]

    [life.streamline_signal(s, gtab, evals=evals) for s in streamline1]

    streamline2 = [[[1, 2, 3], [4, 5, 3], [5, 6, 3], [6, 7, 3]]]

    [life.streamline_signal(s, gtab, evals=evals) for s in streamline2]

    npt.assert_array_equal(streamline2[0], streamline1[0])


def test_voxel2streamline():
    streamline = [
        [[1.1, 2.4, 2.9], [4, 5, 3], [5, 6, 3], [6, 7, 3]],
        [[1, 2, 3], [4, 5, 3], [5, 6, 3]],
    ]
    affine = np.eye(4)
    v2f, v2fn = life.voxel2streamline(streamline, affine)
    npt.assert_equal(v2f, {0: [0, 1], 1: [0, 1], 2: [0, 1], 3: [0]})
    npt.assert_equal(
        v2fn, {0: {0: [0], 1: [1], 2: [2], 3: [3]}, 1: {0: [0], 1: [1], 2: [2]}}
    )
    affine = np.array(
        [[0.9, 0, 0, 10], [0, 0.9, 0, -100], [0, 0, 0.9, 2], [0, 0, 0, 1]]
    )

    xform_sl = life.transform_streamlines(streamline, np.linalg.inv(affine))
    v2f, v2fn = life.voxel2streamline(xform_sl, affine)
    npt.assert_equal(v2f, {0: [0, 1], 1: [0, 1], 2: [0, 1], 3: [0]})
    npt.assert_equal(
        v2fn, {0: {0: [0], 1: [1], 2: [2], 3: [3]}, 1: {0: [0], 1: [1], 2: [2]}}
    )


def test_FiberModel_init():
    # Get some small amount of data:
    data_file, bval_file, bvec_file = dpd.get_fnames(name="small_64D")
    bvals, bvecs = read_bvals_bvecs(bval_file, bvec_file)
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    FM = life.FiberModel(gtab)
    streamline_cases = [
        [
            [[1, 2, 3], [4, 5, 3], [5, 6, 3], [6, 7, 3]],
            [[1, 2, 3], [4, 5, 3], [5, 6, 3]],
        ],
        [[[1, 2, 3]], [[1, 2, 3], [4, 5, 3], [5, 6, 3]]],
    ]

    affine = np.eye(4)

    for sphere in [None, False, dpd.get_sphere(name="symmetric362")]:
        fiber_matrix, vox_coords = FM.setup(streamline_cases[0], affine, sphere=sphere)
        npt.assert_array_equal(
            np.array(vox_coords), np.array([[1, 2, 3], [4, 5, 3], [5, 6, 3], [6, 7, 3]])
        )

        npt.assert_equal(
            fiber_matrix.shape, (len(vox_coords) * 64, len(streamline_cases[0]))
        )
        npt.assert_raises(
            IndexError, FM.setup, streamline_cases[1], affine, sphere=sphere
        )


def test_FiberFit():
    data_file, bval_file, bvec_file = dpd.get_fnames(name="small_64D")
    data = load_nifti_data(data_file)
    bvals, bvecs = read_bvals_bvecs(bval_file, bvec_file)
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    FM = life.FiberModel(gtab)
    evals = [0.0015, 0.0005, 0.0005]

    streamline = [
        [[1, 2, 3], [4, 5, 3], [5, 6, 3], [6, 7, 3]],
        [[1, 2, 3], [4, 5, 3], [5, 6, 3]],
    ]

    fiber_matrix, vox_coords = FM.setup(streamline, np.eye(4), evals=evals)

    w = np.array([0.5, 0.5])
    sig = opt.spdot(fiber_matrix, w) + 1.0  # Add some isotropic stuff
    S0 = data[..., gtab.b0s_mask]
    this_data = np.zeros((10, 10, 10, 64))
    this_data[vox_coords[:, 0], vox_coords[:, 1], vox_coords[:, 2]] = (
        sig.reshape((4, 64)) * S0[vox_coords[:, 0], vox_coords[:, 1], vox_coords[:, 2]]
    )

    # Grab some realistic S0 values:
    this_data = np.concatenate([data[..., gtab.b0s_mask], this_data], -1)

    fit = FM.fit(this_data, streamline, np.eye(4))
    npt.assert_almost_equal(fit.predict()[1], fit.data[1], decimal=-1)

    # Predict with an input GradientTable
    npt.assert_almost_equal(fit.predict(gtab=gtab)[1], fit.data[1], decimal=-1)

    npt.assert_almost_equal(
        this_data[vox_coords[:, 0], vox_coords[:, 1], vox_coords[:, 2]], fit.data
    )


def test_fit_data():
    fdata, fbval, fbvec = dpd.get_fnames(name="small_25")
    fstreamlines = dpd.get_fnames(name="small_25_streamlines")
    gtab = grad.gradient_table(fbval, bvecs=fbvec)
    ni_data = nib.load(fdata)
    data = np.asarray(ni_data.dataobj)

    tensor_streamlines = nib.streamlines.load(fstreamlines).streamlines
    sft = StatefulTractogram(tensor_streamlines, ni_data, Space.RASMM)
    sft.to_vox()
    tensor_streamlines_vox = sft.streamlines

    life_model = life.FiberModel(gtab)
    life_fit = life_model.fit(data, tensor_streamlines_vox, np.eye(4))
    model_error = life_fit.predict() - life_fit.data
    model_rmse = np.sqrt(np.mean(model_error**2, -1))
    matlab_rmse, matlab_weights = dpd.matlab_life_results()
    # Lower error than the matlab implementation for these data:
    npt.assert_(np.median(model_rmse) < np.median(matlab_rmse))
    # And a moderate correlation with the Matlab implementation weights:
    npt.assert_(np.corrcoef(matlab_weights, life_fit.beta)[0, 1] > 0.6)



"""
=========================================================
Connectivity Matrices, ROI Intersections and Density Maps
=========================================================

This example is meant to be an introduction to some of the streamline tools
available in DIPY_. Some of the functions covered in this example are
``target``, ``connectivity_matrix`` and ``density_map``. ``target`` allows one
to filter streamlines that either pass through or do not pass through some
region of the brain, ``connectivity_matrix`` groups and counts streamlines
based on where in the brain they begin and end, and finally, density map counts
the number of streamlines that pass through every voxel of some image.

To get started we'll need to have a set of streamlines to work with. We'll use
EuDX along with the CsaOdfModel to make some streamlines. Let's import the
modules and download the data we'll be using.

Let's load the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import binary_dilation

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.direction import peaks
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst import shm
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.viz import actor, colormap as cmap, window

###############################################################################
# We'll be using the Stanford HARDI dataset which consists of a single
# subject's diffusion, b-values and b-vectors, T1 image and some labels in the
# same space as the T1. We'll use the ``get_fnames`` function to download the
# files we need and set the file names to variables.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
t1_fname = get_fnames(name="stanford_t1")

data, _, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
t1_data = load_nifti_data(t1_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# We've loaded an image called ``labels_img`` which is a map of tissue types
# such that every integer value in the array ``labels`` represents an
# anatomical structure or tissue type [#]_. For this example, the image was
# created so that white matter voxels have values of either 1 or 2. We'll use
# ``peaks_from_model`` to apply the ``CsaOdfModel`` to each white matter voxel
# and estimate fiber orientations which we can use for tracking. We will also
# dilate this mask by 1 voxel to ensure streamlines reach the grey matter.

white_matter = binary_dilation((labels == 1) | (labels == 2))
csamodel = shm.CsaOdfModel(gtab, 6)
csapeaks = peaks.peaks_from_model(
    model=csamodel,
    data=data,
    sphere=peaks.default_sphere,
    relative_peak_threshold=0.8,
    min_separation_angle=45,
    mask=white_matter,
)

###############################################################################
# Now we can use EuDX to track all of the white matter. We define an identity
# matrix for the affine transformation [#]_ of the seeding locations. To keep
# things reasonably fast we use ``density=1`` which will result in 1 seeds per
# voxel. The stopping criterion, determining when the tracking stops, is set to
# stop when the tracking exits the white matter.

affine = np.eye(4)
seeds = utils.seeds_from_mask(white_matter, affine, density=1)
stopping_criterion = BinaryStoppingCriterion(white_matter)

streamline_generator = LocalTracking(
    csapeaks, stopping_criterion, seeds, affine=affine, step_size=0.5
)
streamlines = Streamlines(streamline_generator)

###############################################################################
# The first of the tracking utilities we'll cover here is ``target``. This
# function takes a set of streamlines and a region of interest (ROI) and
# returns only those streamlines that pass through the ROI. The ROI should be
# an array such that the voxels that belong to the ROI are ``True`` and all
# other voxels are ``False`` (this type of binary array is sometimes called a
# mask). This function can also exclude all the streamlines that pass through
# an ROI by setting the ``include`` flag to ``False``. In this example we'll
# target the streamlines of the corpus callosum. Our ``labels`` array has a
# sagittal slice of the corpus callosum identified by the label value 2. We'll
# create an ROI mask from that label and create two sets of streamlines,
# those that intersect with the ROI and those that don't.

cc_slice = labels == 2
cc_streamlines = utils.target(streamlines, affine, cc_slice)
cc_streamlines = Streamlines(cc_streamlines)

other_streamlines = utils.target(streamlines, affine, cc_slice, include=False)
other_streamlines = Streamlines(other_streamlines)
assert len(other_streamlines) + len(cc_streamlines) == len(streamlines)

###############################################################################
# We can use some of DIPY_'s visualization tools to display the ROI we targeted
# above and all the streamlines that pass through that ROI. The ROI is the
# yellow region near the center of the axial image.

# Enables/disables interactive visualization
interactive = False

# Make display objects
color = cmap.line_colors(cc_streamlines)
cc_streamlines_actor = actor.line(
    cc_streamlines, colors=cmap.line_colors(cc_streamlines)
)
cc_ROI_actor = actor.contour_from_roi(cc_slice, color=(1.0, 1.0, 0.0), opacity=0.5)

vol_actor = actor.slicer(t1_data)

vol_actor.display(x=40)
vol_actor2 = vol_actor.copy()
vol_actor2.display(z=35)

# Add display objects to canvas
scene = window.Scene()
scene.add(vol_actor)
scene.add(vol_actor2)
scene.add(cc_streamlines_actor)
scene.add(cc_ROI_actor)

# Save figures
window.record(
    scene=scene, n_frames=1, out_path="corpuscallosum_axial.png", size=(800, 800)
)
if interactive:
    window.show(scene)
scene.set_camera(position=[-1, 0, 0], focal_point=[0, 0, 0], view_up=[0, 0, 1])
window.record(
    scene=scene, n_frames=1, out_path="corpuscallosum_sagittal.png", size=(800, 800)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum Axial and Corpus Callosum Sagittal
#
#
#
# Once we've targeted the corpus callosum ROI, we might want to find out which
# regions of the brain are connected by these streamlines. To do this we can
# use the ``connectivity_matrix`` function. This function takes a set of
# streamlines and an array of labels as arguments. It returns the number of
# streamlines that start and end at each pair of labels and it can return the
# streamlines grouped by their endpoints. Notice that this function only
# considers the endpoints of each streamline.

M, grouping = utils.connectivity_matrix(
    cc_streamlines,
    affine,
    labels.astype(np.uint8),
    return_mapping=True,
    mapping_as_streamlines=True,
)
M[:3, :] = 0
M[:, :3] = 0

###############################################################################
# We've set ``return_mapping`` and ``mapping_as_streamlines`` to ``True`` so
# that ``connectivity_matrix`` returns all the streamlines in
# ``cc_streamlines`` grouped by their endpoint.
#
# Because we're typically only interested in connections between gray matter
# regions, and because the label 0 represents background and the labels 1
# and 2 represent white matter, we discard the first three rows and columns
# of the connectivity matrix.
#
# We can now display this matrix using matplotlib. We display it using a log
# scale to make small values in the matrix easier to see.

plt.imshow(np.log1p(M), interpolation="nearest")
plt.savefig("connectivity.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Connectivity of Corpus Callosum
#
#
#
# In our example track there are more streamlines connecting regions 11 and
# 54 than any other pair of regions. These labels represent the left and right
# superior frontal gyrus respectively. These two regions are large, close
# together, have lots of corpus callosum fibers and are easy to track so this
# result should not be a surprise to anyone.
#
# However, the interpretation of streamline counts can be tricky. The
# relationship between the underlying biology and the streamline counts will
# depend on several factors, including how the tracking was done, and the
# correct way to interpret these kinds of connectivity matrices is still an
# open question in the diffusion imaging literature.
#
# The next function we'll demonstrate is ``density_map``. This function allows
# one to represent the spatial distribution of a track by counting the density
# of streamlines in each voxel. For example, let's take the track connecting
# the left and right superior frontal gyrus.

lr_superiorfrontal_track = grouping[11, 54]
shape = labels.shape
dm = utils.density_map(lr_superiorfrontal_track, affine, shape)

###############################################################################
# Let's save this density map and the streamlines so that they can be
# visualized together. In order to save the streamlines in a ".trk" file we'll
# need to move them to "trackvis space", or the representation of streamlines
# specified by the trackvis Track File format.

# Save density map
save_nifti("lr-superiorfrontal-dm.nii.gz", dm.astype("int16"), affine)

lr_sf_trk = Streamlines(lr_superiorfrontal_track)

# Save streamlines
sft = StatefulTractogram(lr_sf_trk, hardi_img, Space.VOX)
save_trk(sft, "lr-superiorfrontal.trk")

###############################################################################
# .. rubric:: Footnotes
#
# .. [#] The image `aparc-reduced.nii.gz`, which we load as ``labels_img``, is
#        a modified version of label map `aparc+aseg.mgz` created by
#        FreeSurfer_. The corpus  callosum region is a combination of the
#        FreeSurfer labels 251-255. The remaining FreeSurfer labels were
#        re-mapped and reduced so that they lie between 0 and 88. To see the
#        FreeSurfer region, label and name, represented by each value, see
#        `label_info.txt` in `~/.dipy/stanford_hardi`.
# .. [#] An affine transformation is a mapping between two coordinate systems
#        that can represent scaling, rotation, shear, translation and
#        reflection. Affine transformations are often represented using a 4x4
#        matrix where the last row of the matrix is ``[0, 0, 0, 1]``.



"""
=================================================
An introduction to the Deterministic Tractography
=================================================

Deterministic tractography follows
the trajectory of the most probable pathway within the tracking constraint
(e.g. max angle). It follows the direction with the highest
probability from a distribution, as opposed to the probabilistic tractography
which draws the direction from the distribution. Therefore,
deterministic tractography is equivalent to the probabilistic tractography
returning always the maximum value of the distribution.

Deterministic tractography is an alternative to EuDX deterministic
tractography and unlike EuDX does not follow the peaks of the local models but
uses the entire orientation distributions.

This example is an extension of the
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`
example. We begin by loading the data, fitting a Constrained Spherical
Deconvolution (CSD) reconstruction model for the tractography and fitting
the constant solid angle (CSA) reconstruction model to define the tracking
mask (stopping criterion).
"""

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import deterministic_tracking
from dipy.tracking.utils import seeds_from_mask
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False


hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

seed_mask = labels == 2
seeds = seeds_from_mask(seed_mask, affine, density=2)

white_matter = (labels == 1) | (labels == 2)
sc = BinaryStoppingCriterion(white_matter)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
csd_fit = csd_model.fit(data, mask=white_matter)


###############################################################################
# The Fiber Orientation Distribution (FOD) of the CSD model estimates the
# distribution of small fiber bundles within each voxel. This distribution
# can be used for deterministic fiber tracking. As for probabilistic tracking,
# there are many ways to provide those distributions to the deterministic
# tractography. Here, the spherical harmonic representation of
# the FOD is used.

streamline_generator = deterministic_tracking(
    seeds,
    sc,
    affine,
    sh=csd_fit.shm_coeff,
    random_seed=1,
    sphere=default_sphere,
    max_angle=30,
    step_size=0.5,
)

streamlines = Streamlines(streamline_generator)

sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_deterministic.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(scene=scene, out_path="tractogram_deterministic.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using deterministic maximum direction getter



#!/usr/bin/env python3

import logging
import sys

from dipy.data import SPHERE_FILES, get_sphere
from dipy.io.image import load_nifti
from dipy.io.peaks import load_pam
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_tractogram
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import (
    BinaryStoppingCriterion,
    CmcStoppingCriterion,
    ThresholdStoppingCriterion,
)
from dipy.tracking.tracker import (
    closestpeak_tracking,
    deterministic_tracking,
    eudx_tracking,
    pft_tracking,
    probabilistic_tracking,
    ptt_tracking,
)
from dipy.workflows.workflow import Workflow


class LocalFiberTrackingPAMFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "track_local"

    def run(
        self,
        pam_files,
        stopping_files,
        seeding_files,
        use_binary_mask=False,
        stopping_thr=0.2,
        seed_density=1,
        minlen=2,
        maxlen=500,
        step_size=0.5,
        tracking_method="deterministic",
        pmf_threshold=0.1,
        max_angle=30.0,
        sphere_name=None,
        save_seeds=False,
        nbr_threads=0,
        random_seed=1,
        seed_buffer_fraction=1.0,
        out_dir="",
        out_tractogram="tractogram.trk",
    ):
        """Workflow for Local Fiber Tracking.

        This workflow use a saved peaks and metrics (PAM) file as input.

        See :footcite:p:`Garyfallidis2012b` and :footcite:p:`Amirbekian2016`
        for further details about the method.

        Parameters
        ----------
        pam_files : string
           Path to the peaks and metrics files. This path may contain
            wildcards to use multiple masks at once.
        stopping_files : string
            Path to images (e.g. FA) used for stopping criterion for tracking.
        seeding_files : string
            A binary image showing where we need to seed for tracking.
        use_binary_mask : bool, optional
            If True, uses a binary stopping criterion. If the provided
            `stopping_files` are not binary, `stopping_thr` will be used to
            binarize the images.
        stopping_thr : float, optional
            Threshold applied to stopping volume's data to identify where
            tracking has to stop.
        seed_density : int, optional
            Number of seeds per dimension inside voxel.
            For example, seed_density of 2 means 8 regularly distributed
            points in the voxel. And seed density of 1 means 1 point at the
            center of the voxel.
        minlen : int, optional
            Minimum length (nb points) of the streamlines.
        maxlen : int, optional
            Maximum length (nb points) of the streamlines.
        step_size : float, optional
            Step size (in mm) used for tracking.
        tracking_method : string, optional
            Select direction getter strategy:
                - "eudx" (Uses the peaks saved in the pam_files)
                - "deterministic" or "det" for a deterministic tracking
                - "probabilistic" or "prob" for a Probabilistic tracking
                - "closestpeaks" or "cp" for a ClosestPeaks tracking
                - "ptt" for Parallel Transport Tractography

            By default, the sh coefficients saved in the pam_files are used.
        pmf_threshold : float, optional
            Threshold for ODF functions.
        max_angle : float, optional
            Maximum angle between streamline segments (range [0, 90]).
        sphere_name : string, optional
            The sphere used for tracking. If None, the sphere saved in the
            pam_files is used. For faster tracking, use a smaller
            sphere (e.g. 'repulsion200').
        save_seeds : bool, optional
            If true, save the seeds associated to their streamline
            in the 'data_per_streamline' Tractogram dictionary using
            'seeds' as the key.
        nbr_threads : int, optional
            Number of threads to use for the processing. By default, all available
            threads will be used.
        random_seed : int, optional
            Seed for the random number generator, must be >= 0. A value of greater
            than 0 will all produce the same streamline trajectory for a given seed
            coordinate. A value of 0 may produces various streamline tracjectories
            for a given seed coordinate.
        seed_buffer_fraction : float, optional
            Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
            buffer. A value of 0.5 will use half of the seed buffer then the other half.
            a way to reduce memory usage.
        out_dir : string, optional
           Output directory.
        out_tractogram : string, optional
           Name of the tractogram file to be saved.

        References
        ----------
        .. footbibliography::

        """
        io_it = self.get_io_iterator()

        tracking_method = tracking_method.lower()
        sphere_name = SPHERE_FILES.get(sphere_name, None)

        for pams_path, stopping_path, seeding_path, out_tract in io_it:
            if tracking_method == "eudx":
                logging.info(f"EuDX Deterministic tracking on {pams_path}")
            else:
                logging.info(f"{tracking_method.title()} tracking on {pams_path}")

            pam = load_pam(pams_path, verbose=False)
            sphere = pam.sphere if sphere_name is None else get_sphere(name=sphere_name)

            logging.info("Loading stopping criterion")
            stop, affine = load_nifti(stopping_path)
            if use_binary_mask:
                stopping_criterion = BinaryStoppingCriterion(stop > stopping_thr)
            else:
                stopping_criterion = ThresholdStoppingCriterion(stop, stopping_thr)

            logging.info("Loading seeds")
            seed_mask, _ = load_nifti(seeding_path)
            seeds = utils.seeds_from_mask(
                seed_mask, affine, density=[seed_density, seed_density, seed_density]
            )

            logging.info("Starting to track")
            if tracking_method in ["closestpeaks", "cp"]:
                tracking_result = closestpeak_tracking(
                    seeds,
                    stopping_criterion,
                    affine,
                    sh=pam.shm_coeff,
                    random_seed=random_seed,
                    sphere=sphere,
                    max_angle=max_angle,
                    min_len=minlen,
                    max_len=maxlen,
                    step_size=step_size,
                    pmf_threshold=pmf_threshold,
                    save_seeds=save_seeds,
                    nbr_threads=nbr_threads,
                    seed_buffer_fraction=seed_buffer_fraction,
                )
            elif tracking_method in [
                "eudx",
            ]:
                tracking_result = eudx_tracking(
                    seeds,
                    stopping_criterion,
                    affine,
                    sh=pam.shm_coeff,
                    pam=pam,
                    random_seed=random_seed,
                    sphere=sphere,
                    max_angle=max_angle,
                    min_len=minlen,
                    max_len=maxlen,
                    step_size=step_size,
                    pmf_threshold=pmf_threshold,
                    save_seeds=save_seeds,
                    nbr_threads=nbr_threads,
                    seed_buffer_fraction=seed_buffer_fraction,
                )
            elif tracking_method in ["deterministic", "det"]:
                tracking_result = deterministic_tracking(
                    seeds,
                    stopping_criterion,
                    affine,
                    sh=pam.shm_coeff,
                    random_seed=random_seed,
                    sphere=sphere,
                    max_angle=max_angle,
                    min_len=minlen,
                    max_len=maxlen,
                    step_size=step_size,
                    pmf_threshold=pmf_threshold,
                    save_seeds=save_seeds,
                    nbr_threads=nbr_threads,
                    seed_buffer_fraction=seed_buffer_fraction,
                )
            elif tracking_method in ["probabilistic", "prob"]:
                tracking_result = probabilistic_tracking(
                    seeds,
                    stopping_criterion,
                    affine,
                    sh=pam.shm_coeff,
                    random_seed=random_seed,
                    sphere=sphere,
                    max_angle=max_angle,
                    min_len=minlen,
                    max_len=maxlen,
                    step_size=step_size,
                    pmf_threshold=pmf_threshold,
                    save_seeds=save_seeds,
                    nbr_threads=nbr_threads,
                    seed_buffer_fraction=seed_buffer_fraction,
                )
            elif tracking_method in ["ptt"]:
                tracking_result = ptt_tracking(
                    seeds,
                    stopping_criterion,
                    affine,
                    sh=pam.shm_coeff,
                    random_seed=random_seed,
                    sphere=sphere,
                    max_angle=max_angle,
                    min_len=minlen,
                    max_len=maxlen,
                    step_size=step_size,
                    pmf_threshold=pmf_threshold,
                    save_seeds=save_seeds,
                    nbr_threads=nbr_threads,
                    seed_buffer_fraction=seed_buffer_fraction,
                )
            else:
                logging.error(
                    f"Unknown tracking method: {tracking_method}. "
                    f"Please use one of the following: "
                    f"'eudx', 'deterministic', 'probabilistic', 'closestpeaks', 'ptt'"
                )
                sys.exit(1)
            if save_seeds:
                streamlines, seeds = zip(*tracking_result)
                seeds = {"seeds": seeds}
            else:
                streamlines = list(tracking_result)
                seeds = {}

            sft = StatefulTractogram(
                streamlines, seeding_path, Space.RASMM, data_per_streamline=seeds
            )
            save_tractogram(sft, out_tract, bbox_valid_check=False)
            logging.info(f"Saved {out_tract}")


class PFTrackingPAMFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "track_pft"

    def run(
        self,
        pam_files,
        wm_files,
        gm_files,
        csf_files,
        seeding_files,
        step_size=0.2,
        seed_density=1,
        pmf_threshold=0.1,
        max_angle=20.0,
        sphere_name=None,
        pft_back=2,
        pft_front=1,
        pft_count=15,
        pft_max_trial=20,
        save_seeds=False,
        min_wm_pve_before_stopping=0,
        nbr_threads=0,
        random_seed=1,
        seed_buffer_fraction=1.0,
        out_dir="",
        out_tractogram="tractogram.trk",
    ):
        """Workflow for Particle Filtering Tracking.

        This workflow uses a saved peaks and metrics (PAM) file as input.

        See :footcite:p:`Girard2014` for further details about the method.

        Parameters
        ----------
        pam_files : string
           Path to the peaks and metrics files. This path may contain
            wildcards to use multiple masks at once.
        wm_files : string
            Path to white matter partial volume estimate for tracking (CMC).
        gm_files : string
            Path to grey matter partial volume estimate for tracking (CMC).
        csf_files : string
            Path to cerebrospinal fluid partial volume estimate for tracking
            (CMC).
        seeding_files : string
            A binary image showing where we need to seed for tracking.
        step_size : float, optional
            Step size (in mm) used for tracking.
        seed_density : int, optional
            Number of seeds per dimension inside voxel.
            For example, seed_density of 2 means 8 regularly distributed
            points in the voxel. And seed density of 1 means 1 point at the
            center of the voxel.
        pmf_threshold : float, optional
            Threshold for ODF functions.
        max_angle : float, optional
            Maximum angle between streamline segments (range [0, 90]).
        sphere_name : string, optional
            The sphere used for tracking. If None, the sphere saved in the
            pam_files is used. For faster tracking, use a smaller
            sphere (e.g. 'repulsion200').
        pft_back : float, optional
            Distance in mm to back track before starting the particle filtering
            tractography. The total particle filtering
            tractography distance is equal to back_tracking_dist +
            front_tracking_dist.
        pft_front : float, optional
            Distance in mm to run the particle filtering tractography after the
            the back track distance. The total particle filtering
            tractography distance is equal to back_tracking_dist +
            front_tracking_dist.
        pft_count : int, optional
            Number of particles to use in the particle filter.
        pft_max_trial : int, optional
            Maximum number of trials to run the particle filtering tractography.
        save_seeds : bool, optional
            If true, save the seeds associated to their streamline
            in the 'data_per_streamline' Tractogram dictionary using
            'seeds' as the key.
        min_wm_pve_before_stopping : int, optional
            Minimum white matter pve (1 - stopping_criterion.include_map -
            stopping_criterion.exclude_map) to reach before allowing the
            tractography to stop.
        nbr_threads : int, optional
            Number of threads to use for the processing. By default, all available
            threads will be used.
        random_seed : int, optional
            Seed for the random number generator, must be >= 0. A value of greater
            than 0 will all produce the same streamline trajectory for a given seed
            coordinate. A value of 0 may produces various streamline tracjectories
            for a given seed coordinate.
        seed_buffer_fraction : float, optional
            Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
            buffer. A value of 0.5 will use half of the seed buffer then the other half.
            a way to reduce memory usage.
        out_dir : string, optional
           Output directory.
        out_tractogram : string, optional
           Name of the tractogram file to be saved.

        References
        ----------
        .. footbibliography::

        """
        io_it = self.get_io_iterator()
        sphere_name = SPHERE_FILES.get(sphere_name, None)

        for pams_path, wm_path, gm_path, csf_path, seeding_path, out_tract in io_it:
            logging.info(f"Particle Filtering tracking on {pams_path}")

            pam = load_pam(pams_path, verbose=False)
            sphere = pam.sphere if sphere_name is None else get_sphere(name=sphere_name)

            wm, affine, voxel_size = load_nifti(wm_path, return_voxsize=True)
            gm, _ = load_nifti(gm_path)
            csf, _ = load_nifti(csf_path)
            avs = sum(voxel_size) / len(voxel_size)  # average_voxel_size

            logging.info("Preparing stopping criterion")
            stopping_criterion = CmcStoppingCriterion.from_pve(
                wm, gm, csf, step_size=step_size, average_voxel_size=avs
            )

            logging.info("Seeding in mask")
            seed_mask, _ = load_nifti(seeding_path)
            seeds = utils.seeds_from_mask(
                seed_mask, affine, density=[seed_density, seed_density, seed_density]
            )

            logging.info("Start tracking")
            tracking_result = pft_tracking(
                seeds,
                stopping_criterion,
                affine,
                sh=pam.shm_coeff,
                sphere=sphere,
                step_size=step_size,
                pft_back_tracking_dist=pft_back,
                pft_front_tracking_dist=pft_front,
                pft_max_trial=pft_max_trial,
                particle_count=pft_count,
                save_seeds=save_seeds,
                min_wm_pve_before_stopping=min_wm_pve_before_stopping,
                random_seed=random_seed,
                max_angle=max_angle,
                pmf_threshold=pmf_threshold,
                nbr_threads=nbr_threads,
                seed_buffer_fraction=seed_buffer_fraction,
            )

            if save_seeds:
                streamlines, seeds = zip(*tracking_result)
                seeds = {"seeds": seeds}
            else:
                streamlines = list(tracking_result)
                seeds = {}

            sft = StatefulTractogram(
                streamlines, seeding_path, Space.RASMM, data_per_streamline=seeds
            )
            save_tractogram(sft, out_tract, bbox_valid_check=False)
            logging.info(f"Saved {out_tract}")



"""
======================
Fast Streamline Search
======================

This example explains how Fast Streamline Search :footcite:p:`StOnge2022`
can be used to find all similar streamlines.

First import the necessary modules.
"""

import numpy as np

from dipy.data import (
    fetch_bundle_atlas_hcp842,
    fetch_target_tractogram_hcp,
    get_target_tractogram_hcp,
    get_two_hcp842_bundles,
)
from dipy.io.streamline import load_trk
from dipy.segment.fss import FastStreamlineSearch, nearest_from_matrix_row
from dipy.viz import actor, window

###############################################################################
# Download and read data for this tutorial

fetch_bundle_atlas_hcp842()
fetch_target_tractogram_hcp()

hcp_file = get_target_tractogram_hcp()
streamlines = load_trk(hcp_file, "same", bbox_valid_check=False).streamlines

###############################################################################
# Visualize the atlas (ref) bundle and full brain tractogram

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(streamlines))
if interactive:
    window.show(scene)
else:
    window.record(scene=scene, out_path="tractograms_initial.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Atlas bundle and source streamlines before registration.
#
#
#
# Read Arcuate Fasciculus Left and Corticospinal Tract Left bundles from
# already fetched atlas data to use them as model bundle. Let's visualize the
# Arcuate Fasciculus Left model bundle.

model_af_l_file, model_cst_l_file = get_two_hcp842_bundles()
sft_af_l = load_trk(model_af_l_file, "same", bbox_valid_check=False)
model_af_l = sft_af_l.streamlines

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_af_l, colors=(0, 1, 0)))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
if interactive:
    window.show(scene)
else:
    window.record(scene=scene, out_path="AF_L_model_bundle.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Model Arcuate Fasciculus Left bundle
#
#
#
#
# Search for all similar streamlines :footcite:p:`StOnge2022`
#
# Fast Streamline Search can do a radius search to find all streamlines that
# are similar to from one tractogram to another. It returns the distance
# matrix mapping between both tractograms. The same list of streamlines can
# be given to recover the self distance matrix.
#
# here are the ``FastStreamlinesSearch`` class need the following
# initialization arguments:
#
# - ref_streamlines : reference streamlines, that will be searched in (tree)
# - max_radius : is the maximum distance that can be used with radius search
#
# Then, the ``radius_search()`` method needs the following arguments:
#
# - radius : for each streamline search find all similar ones in the
#   "ref_streamlines" that are within the given radius
#
# Be cautious, a large radius might result in a dense distance computation,
# requiring a large amount of time and memory. Recommended range of the
# radius is from 1 to 10 mm.

radius = 7.0
fs_tree_af = FastStreamlineSearch(ref_streamlines=model_af_l, max_radius=radius)
coo_mdist_mtx = fs_tree_af.radius_search(streamlines, radius=radius)

###############################################################################
# Extract indices of streamlines with an similar ones in the reference

ids_s = np.unique(coo_mdist_mtx.row)
ids_ref = np.unique(coo_mdist_mtx.col)

recognized_af_l = streamlines[ids_s].copy()

###############################################################################
# let's visualize streamlines similar to the Arcuate Fasciculus Left bundle

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_af_l, colors=(0, 0, 1)))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
if interactive:
    window.show(scene)
else:
    window.record(scene=scene, out_path="AF_L_recognized_bundle.png", size=(600, 600))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Recognized Arcuate Fasciculus Left bundle
#
#
#
#
# Color the resulting AF by the nearest streamlines distance

nn_s, nn_ref, nn_dist = nearest_from_matrix_row(coo_mdist_mtx)

scene = window.Scene()
scene.SetBackground(1, 1, 1)
cmap = actor.colormap_lookup_table(scale_range=(nn_dist.min(), nn_dist.max()))
scene.add(actor.line(recognized_af_l, colors=nn_dist, lookup_colormap=cmap))
scene.add(actor.scalar_bar(lookup_table=cmap, title="distance to atlas (mm)"))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
if interactive:
    window.show(scene)
else:
    window.record(
        scene=scene, out_path="AF_L_recognized_bundle_dist.png", size=(600, 600)
    )

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Recognized Arcuate Fasciculus Left bundle colored by distance to ref
#
#
#
# Display the streamlines reference reached in green

# Default red color
ref_color = np.zeros((len(model_af_l), 3), dtype=float)
ref_color[:] = (1.0, 0.0, 0.0)

# Reached in green color
ref_color[ids_ref] = (0.0, 1.0, 0.0)

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_af_l, colors=ref_color))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)

if interactive:
    window.show(scene)
else:
    window.record(
        scene=scene, out_path="AF_L_model_bundle_reached.png", size=(600, 600)
    )

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Arcuate Fasciculus Left model reached (green) in radius
#
#
#
#
# References
# ----------
#
# .. footbibliography::
#



"""Tools to easily make multi voxel models"""

from functools import partial
import multiprocessing

import numpy as np
from tqdm import tqdm

from dipy.core.ndindex import ndindex
from dipy.reconst.base import ReconstFit
from dipy.reconst.quick_squash import quick_squash as _squash
from dipy.utils.parallel import paramap


def _parallel_fit_worker(vox_data, single_voxel_fit, **kwargs):
    """
    Works on a chunk of voxel data to create a list of
    single voxel fits.

    Parameters
    ----------
    vox_data : ndarray, shape (n_voxels, ...)
        The data to fit.

    single_voxel_fit : callable
        The fit function to use on each voxel.
    """
    vox_weights = kwargs.pop("weights", None)
    if type(vox_weights) is np.ndarray:
        return [
            single_voxel_fit(data, **(dict({"weights": weights}, **kwargs)))
            for data, weights in zip(vox_data, vox_weights)
        ]
    else:
        return [single_voxel_fit(data, **kwargs) for data in vox_data]


def multi_voxel_fit(single_voxel_fit):
    """Method decorator to turn a single voxel model fit
    definition into a multi voxel model fit definition
    """

    def new_fit(self, data, *, mask=None, **kwargs):
        """Fit method for every voxel in data"""

        # If only one voxel just return a standard fit, passing through
        # the functions key-word arguments (no mask needed).
        if data.ndim == 1:
            svf = single_voxel_fit(self, data, **kwargs)
            # If fit method does not return extra, cannot return extra
            if isinstance(svf, tuple):
                svf, extra = svf
                return svf, extra
            else:
                return svf

        # Make a mask if mask is None
        if mask is None:
            mask = np.ones(data.shape[:-1], bool)
        # Check the shape of the mask if mask is not None
        elif mask.shape != data.shape[:-1]:
            raise ValueError("mask and data shape do not match")

        # Get weights from kwargs if provided
        weights = kwargs["weights"] if "weights" in kwargs else None
        weights_is_array = True if type(weights) is np.ndarray else False

        # Fit data where mask is True
        fit_array = np.empty(data.shape[:-1], dtype=object)
        return_extra = False

        # Default to serial execution:
        engine = kwargs.get("engine", "serial")
        if engine == "serial":
            extra_list = []
            bar = tqdm(
                total=np.sum(mask), position=0, disable=kwargs.get("verbose", True)
            )
            bar.set_description("Fitting reconstruction model using serial execution")
            for ijk in ndindex(data.shape[:-1]):
                if mask[ijk]:
                    if weights_is_array:
                        kwargs["weights"] = weights[ijk]

                    svf = single_voxel_fit(self, data[ijk], **kwargs)

                    # Not all fit methods return extra, handle this here
                    if isinstance(svf, tuple):
                        fit_array[ijk], extra = svf
                        return_extra = True
                    else:
                        fit_array[ijk], extra = svf, None

                    extra_list.append(extra)

                bar.update()
            bar.close()
        else:
            data_to_fit = data[np.where(mask)]
            if weights_is_array:
                weights_to_fit = weights[np.where(mask)]
            single_voxel_with_self = partial(single_voxel_fit, self)
            n_jobs = kwargs.get("n_jobs", multiprocessing.cpu_count() - 1)
            vox_per_chunk = kwargs.get(
                "vox_per_chunk", np.max([data_to_fit.shape[0] // n_jobs, 1])
            )
            chunks = [
                data_to_fit[ii : ii + vox_per_chunk]
                for ii in range(0, data_to_fit.shape[0], vox_per_chunk)
            ]

            # func_kwargs : dict or sequence, optional
            # Keyword arguments to `func` or sequence of keyword arguments
            # to `func`: one item for each item in the input list.
            kwargs_chunks = []
            for ii in range(0, data_to_fit.shape[0], vox_per_chunk):
                kw = kwargs.copy()
                if weights_is_array:
                    kw["weights"] = weights_to_fit[ii : ii + vox_per_chunk]
                kwargs_chunks.append(kw)

            parallel_kwargs = {}
            for kk in ["n_jobs", "vox_per_chunk", "engine", "verbose"]:
                if kk in kwargs:
                    parallel_kwargs[kk] = kwargs[kk]

            mvf = paramap(
                _parallel_fit_worker,
                chunks,
                func_args=[single_voxel_with_self],
                func_kwargs=kwargs_chunks,
                **parallel_kwargs,
            )

            if isinstance(mvf[0][0], tuple):
                tmp_fit_array = np.concatenate(
                    [[svf[0] for svf in mvf_ch] for mvf_ch in mvf]
                )
                tmp_extra = np.concatenate(
                    [[svf[1] for svf in mvf_ch] for mvf_ch in mvf]
                ).tolist()
                fit_array[np.where(mask)], extra_list = tmp_fit_array, tmp_extra
                return_extra = True
            else:
                tmp_fit_array = np.concatenate(mvf)
                fit_array[np.where(mask)], extra_list = tmp_fit_array, None

        # Redefine extra to be a single dictionary
        if return_extra:
            if extra_list[0] is not None:
                extra_mask = {
                    key: np.vstack([e[key] for e in extra_list])
                    for key in extra_list[0]
                }
                extra = {}
                for key in extra_mask:
                    extra[key] = np.zeros(data.shape)
                    extra[key][mask == 1] = extra_mask[key]
            else:
                extra = None

        # If fit method does not return extra, assume we cannot return extra
        if return_extra:
            return MultiVoxelFit(self, fit_array, mask), extra
        else:
            return MultiVoxelFit(self, fit_array, mask)

    return new_fit


class MultiVoxelFit(ReconstFit):
    """Holds an array of fits and allows access to their attributes and
    methods"""

    def __init__(self, model, fit_array, mask):
        self.model = model
        self.fit_array = fit_array
        self.mask = mask

    @property
    def shape(self):
        return self.fit_array.shape

    def __getattr__(self, attr):
        result = CallableArray(self.fit_array.shape, dtype=object)
        for ijk in ndindex(result.shape):
            if self.mask[ijk]:
                result[ijk] = getattr(self.fit_array[ijk], attr)
        return _squash(result, self.mask)

    def __getitem__(self, index):
        item = self.fit_array[index]
        if isinstance(item, np.ndarray):
            return MultiVoxelFit(self.model, item, self.mask[index])
        else:
            return item

    def predict(self, *args, **kwargs):
        """
        Predict for the multi-voxel object using each single-object's
        prediction API, with S0 provided from an array.
        """
        S0 = kwargs.get("S0", np.ones(self.fit_array.shape))
        idx = ndindex(self.fit_array.shape)
        ijk = next(idx)

        def gimme_S0(S0, ijk):
            if isinstance(S0, np.ndarray):
                return S0[ijk]
            else:
                return S0

        kwargs["S0"] = gimme_S0(S0, ijk)
        # If we have a mask, we might have some Nones up front, skip those:
        while self.fit_array[ijk] is None:
            ijk = next(idx)

        if not hasattr(self.fit_array[ijk], "predict"):
            msg = "This model does not have prediction implemented yet"
            raise NotImplementedError(msg)

        first_pred = self.fit_array[ijk].predict(*args, **kwargs)
        result = np.zeros(self.fit_array.shape + (first_pred.shape[-1],))
        result[ijk] = first_pred
        for ijk in idx:
            kwargs["S0"] = gimme_S0(S0, ijk)
            # If it's masked, we predict a 0:
            if self.fit_array[ijk] is None:
                result[ijk] *= 0
            else:
                result[ijk] = self.fit_array[ijk].predict(*args, **kwargs)

        return result


class CallableArray(np.ndarray):
    """An array which can be called like a function"""

    def __call__(self, *args, **kwargs):
        result = np.empty(self.shape, dtype=object)
        for ijk in ndindex(self.shape):
            item = self[ijk]
            if item is not None:
                result[ijk] = item(*args, **kwargs)
        return _squash(result)



from dipy.viz.horizon.tab import HorizonTab, build_checkbox, build_slider


class ROIsTab(HorizonTab):
    def __init__(self, contour_actors):
        """Initialize interaction tab for ROIs visualization.

        Parameters
        ----------
        contour_actors : list
            list of vtkActor.
        """

        super().__init__()

        self._actors = contour_actors
        self._name = "ROIs"

        self._tab_id = 0

        self._actor_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._toggle_actors
        )

        self._opacity_slider_label, self._opacity_slider = build_slider(
            initial_value=1,
            max_value=1.0,
            text_template="{ratio:.0%}",
            on_change=self._change_opacity,
            label="Opacity",
        )

        self._register_elements(
            self._actor_toggle, self._opacity_slider_label, self._opacity_slider
        )

    def _change_opacity(self, slider):
        """Change opacity of all ROIs.

        Parameters
        ----------
        slider : LineSlider2D
        """
        opacity = slider.value
        self._opacity_slider.selected_value = slider.value
        for contour in self._actors:
            contour.GetProperty().SetOpacity(opacity)

    def build(self, tab_id):
        """Position the elements in the tab.

        Parameters
        ----------
        tab_id : int
            Identifier for the tab. Index of the tab in TabUI.
        """

        self._tab_id = tab_id

        y_pos = 0.85
        self._actor_toggle.position = (0.02, y_pos)
        self._opacity_slider_label.position = (0.05, y_pos)
        self._opacity_slider.position = (0.12, y_pos)

    @property
    def name(self):
        """Title of the tab.

        Returns
        -------
        str
        """
        return self._name

    @property
    def actors(self):
        """Actors controlled by tab.

        Returns
        -------
        list
        """
        return self._actors



"""
=====================================================================
Applying positivity constraints to Q-space Trajectory Imaging (QTI+)
=====================================================================

Q-space trajectory imaging (QTI) :footcite:p:`Westin2016` with applied
positivity constraints (QTI+) is an estimation framework proposed by
:footcite:t:`Herberthson2021` which enforces necessary constraints during the
estimation of the QTI model parameters.

This tutorial briefly summarizes the theory and provides a comparison between
performing the constrained and unconstrained QTI reconstruction in DIPY.

Theory
======

In QTI, the tissue microstructure is represented by a diffusion tensor
distribution (DTD). Here, DTD is denoted by $\\mathbf{D}$ and the voxel-level
diffusion tensor from DTI by $\\langle\\mathbf{D}\\rangle$, where
$\\langle \\ \\rangle$ denotes averaging over the DTD. The covariance of
$\\mathbf{D}$ is given by a fourth-order covariance tensor $\\mathbb{C}$
defined as

.. math::

   \\mathbb{C} = \\langle \\mathbf{D} \\otimes \\mathbf{D} \\rangle - \\langle
   \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D} \\rangle ,

where $\\otimes$ denotes a tensor outer product. $\\mathbb{C}$ has 21 unique
elements and enables the calculation of several microstructural parameters.

Using the cumulant expansion, the diffusion-weighted signal can be approximated
as

.. math::

   S \\approx S_0 \\exp \\left(- \\mathbf{b} : \\langle \\mathbf{D} \\rangle +
   \\frac{1}{2}(\\mathbf{b} \\otimes \\mathbf{b}) : \\mathbb{C} \\right) ,

where $S_0$ is the signal without diffusion-weighting, $\\mathbf{b}$ is the
b-tensor used in the acquisition, and $:$ denotes a tensor inner product.

The model parameters $S_0$, $\\langle \\mathbf{D}\\rangle$, and $\\mathbb{C}$
can be estimated by solving the following weighted problem, where the
heteroskedasticity introduced by the taking the logarithm of the signal is
accounted for:

.. math::

   {\\mathrm{argmin}}_{S_0,\\langle \\mathbf{D} \\rangle, \\mathbb{C}}
   \\sum_{k=1}^n S_k^2 \\left| \\ln(S_k)-\\ln(S_0)+\\mathbf{b}^{(k)} \\langle
   \\mathbf{D} \\rangle -\\frac{1}{2} (\\mathbf{b} \\otimes \\mathbf{b})^{(k)}
   \\mathbb{C} \\right|^2 ,

the above can be written as a weighted least squares problem

.. math::

   \\mathbf{Ax} = \\mathbf{y},

where

.. math::

   y = \\begin{pmatrix} \\ S_1 \\ln S_1 \\\\ \\vdots \\\\
   \\ S_n \\ln S_n \\end{pmatrix} ,

.. math::

   x = \\begin{pmatrix} \\ln S_0 & \\langle \\mathbf{D} \\rangle & \\mathbb{C}
   \\end{pmatrix}^\\text{T} ,

.. math::

   A =
   \\begin{pmatrix}
   S_1 & 0 & \\ldots & 0 \\\\ 0 & \\ddots & \\ddots & \\vdots \\\\ \\vdots &
   \\ddots & \\ddots & 0 \\\\ 0 & \\ldots & 0 & S_n
   \\end{pmatrix}
   \\begin{pmatrix}
   1 & -\\mathbf{b}_1^\\text{T} & \\frac{1}{2} (\\mathbf{b}_1 \\otimes
   \\mathbf{b}_1) \\text{T} \\\\
   \\vdots & \\vdots & \\vdots \\\\
   \\vdots & \\vdots & \\vdots \\\\
   1 & -\\mathbf{b}_n^\\text{T} & \\frac{1}{2} (\\mathbf{b}_n \\otimes
   \\mathbf{b}_n) ^\\text{T}
   \\end{pmatrix} ,

where $n$ is the number of acquisitions and $\\langle\\mathbf{D}\\rangle$,
$\\mathbb{C}$, $\\mathbf{b}_i$, and $(\\mathbf{b}_i \\otimes \\mathbf{b}_i)$
are represented by column vectors using Voigt notation.

The estimated $\\langle\\mathbf{D}\\rangle$ and $\\mathbb{C}$ tensors
should observe mathematical and physical conditions dictated by the model
itself: since $\\langle\\mathbf{D}\\rangle$ represents a diffusivity, it
should be positive semi-definite: $\\langle\\mathbf{D}\\rangle \\succeq 0$.
Similarly, since $\\mathbf{C}$ represents a covariance, it's $6 \\times 6$
representation, $\\mathbf{C}$, should be positive semi-definite:
$\\mathbf{C} \\succeq 0$

When not imposed, violations of these conditions can occur in presence of noise
and/or in sparsely sampled data. This could result in metrics derived from the
model parameters to be unreliable. Both these conditions can be enfoced while
estimating the QTI model's parameters using semidefinite programming (SDP) as
shown by :footcite:t:`Herberthson2021`. This corresponds to solving the problem

.. math::

    \\mathbf{Ax} = \\mathbf{y}
    \\text{subject to:}
    \\langle\\mathbf{D}\\rangle \\succeq 0 ,
    \\mathbf{C} \\succeq 0

Installation
=============

The constrained problem stated above can be solved using the cvxpy library.
Instructions on how to install cvxpy
can be found at https://www.cvxpy.org/install/. A free SDP solver
called 'SCS' is installed with cvxpy, and can be readily used for
performing the fit. However, it is recommended to install an
alternative solver, MOSEK, for improved speed and performance.
MOSEK requires a licence which is free for academic use.
Instructions on how to install Mosek and setting up a licence can be found
at https://docs.mosek.com/latest/install/installation.html

Usage example
==============

Here we show how metrics derived from the
QTI model parameters compare when the fit is performed with and without
applying the positivity constraints.

In DIPY, the constrained estimation routine is available as part of the
`dipy.reconst.qti` module.
First we import all the necessary modules to perform the QTI fit:
"""

from dipy.data import read_DiB_70_lte_pte_ste, read_DiB_217_lte_pte_ste
import dipy.reconst.qti as qti
from dipy.viz.plotting import compare_qti_maps

###############################################################################
# To showcase why enforcing positivity constraints in QTI is relevant, we use
# a human brain dataset comprising 70 volumes acquired with tensor-encoding.
# This dataset was obtained by subsampling a richer dataset containing 217
# diffusion measurements, which we will use as ground truth when comparing
# the parameters estimation with and without applied constraints. This emulates
# performing shorter data acquisition which can correspond to scanning patients
# in clinical settings.
#
# The full dataset used in this tutorial was originally published at
# https://github.com/filip-szczepankiewicz/Szczepankiewicz_DIB_2019,
# and is described in :footcite:p:`Szczepankiewicz2019`.
#
#
# First, let's load the complete dataset and create the gradient table.
# We mark these two with the '_217' suffix.

data_img, mask_img, gtab_217 = read_DiB_217_lte_pte_ste()
data_217 = data_img.get_fdata()
mask = mask_img.get_fdata()

###############################################################################
# Second, let's load the downsampled dataset and create the gradient table.
# We mark these two with the '_70' suffix.

data_img, _, gtab_70 = read_DiB_70_lte_pte_ste()
data_70 = data_img.get_fdata()

###############################################################################
# Now we can fit the QTI model to the datasets containing 217 measurements, and
# use it as reference to compare the constrained and unconstrained fit on the
# smaller dataset. For time considerations, we restrict the fit to a
# single slice.

mask[:, :, :13] = 0
mask[:, :, 14:] = 0

qtimodel_217 = qti.QtiModel(gtab_217)
qtifit_217 = qtimodel_217.fit(data_217, mask=mask)

###############################################################################
# Now we can fit the QTI model using the default unconstrained fitting method
# to the subsampled dataset:

qtimodel_unconstrained = qti.QtiModel(gtab_70)
qtifit_unconstrained = qtimodel_unconstrained.fit(data_70, mask=mask)

###############################################################################
# Now we repeat the fit but with the constraints applied.
# To perform the constrained fit, we select the 'SDPdc' fit method when
# creating the QtiModel object.
#
# .. note::
#     this fit method is slower compared to the defaults unconstrained.
#
# If mosek is installed, it can be specified as the solver to be used
# as follows:
#
# .. code-block:: python
#
#     qtimodel = qti.QtiModel(gtab, fit_method='SDPdc', cvxpy_solver='MOSEK')
#     qtifit = qtimodel.fit(data, mask)
#
# If Mosek is not installed, the constrained fit can still be performed, and
# SCS will be used as solver. SCS is typically much slower than Mosek, but
# provides similar results in terms of accuracy. To give an example, the fit
# performed in the next line will take approximately 15 minutes when using SCS,
# and 2 minute when using Mosek!

qtimodel_constrained = qti.QtiModel(gtab_70, fit_method="SDPdc")
qtifit_constrained = qtimodel_constrained.fit(data_70, mask=mask)

###############################################################################
# Now we can visualize the results obtained with the constrained and
# unconstrained fit on the small dataset, and compare them with the
# "ground truth" provided by fitting the QTI model to the full dataset.
# For example, we can look at the FA and µFA maps, and their value
# distribution in White Matter in comparison to the ground truth.

z = 13
wm_mask = qtifit_217.ufa[:, :, z] > 0.6
compare_qti_maps(qtifit_217, qtifit_unconstrained, qtifit_constrained, wm_mask)

###############################################################################
# The results clearly show how many of the FA and µFA values
# obtained with the unconstrained fit fall outside the correct
# theoretical range [0 1], while the constrained fit provides
# more sound results. Note also that even when fitting the rich
# dataset, some values of the parameters produced with the unconstrained
# fit fall outside the correct range, suggesting that the constrained fit,
# despite the additional time cost, should be performed even on densely
# sampled diffusion data.
#
# For more information about QTI and QTI+, please read the articles by
# :footcite:p:`Westin2016` and :footcite:t:`Herberthson2021`.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
=================================================
Reconstruct with Q-space Trajectory Imaging (QTI)
=================================================

Q-space trajectory imaging (QTI) by :footcite:t:`Westin2016` is a general
framework for analyzing diffusion-weighted MRI data acquired with tensor-valued
diffusion encoding. This tutorial briefly summarizes the theory and provides an
example of how to estimate the diffusion and covariance tensors using DIPY.

Theory
======

In QTI, the tissue microstructure is represented by a diffusion tensor
distribution (DTD). Here, DTD is denoted by $\\mathbf{D}$ and the voxel-level
diffusion tensor from DTI by $\\langle\\mathbf{D}\\rangle$, where
$\\langle \\ \\rangle$ denotes averaging over the DTD. The covariance of
$\\mathbf{D}$ is given by a fourth-order covariance tensor $\\mathbb{C}$
defined as

.. math::

   \\mathbb{C} = \\langle \\mathbf{D} \\otimes \\mathbf{D} \\rangle - \\langle
   \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D} \\rangle ,

where $\\otimes$ denotes a tensor outer product. $\\mathbb{C}$ has 21 unique
elements and enables the calculation of several microstructural parameters.

Using the cumulant expansion, the diffusion-weighted signal can be approximated
as

.. math::
   S \\approx S_0 \\exp \\left(- \\mathbf{b} : \\langle \\mathbf{D} \\rangle +
   \\frac{1}{2}(\\mathbf{b} \\otimes \\mathbf{b}) : \\mathbb{C} \\right) ,

where $S_0$ is the signal without diffusion-weighting, $\\mathbf{b}$ is the
b-tensor used in the acquisition, and $:$ denotes a tensor inner product.

The model parameters $S_0$, $\\langle\\mathbf{D}\\rangle$, and $\\mathbb{C}$
can be estimated by solving the following equation:

.. math::

   S = \\beta X,

where

.. math::

   S = \\begin{pmatrix} \\ln S_1 \\\\ \\vdots \\\\ \\ln S_n \\end{pmatrix} ,

.. math::

   \\beta = \\begin{pmatrix} \\ln S_0 & \\langle \\mathbf{D}
   \\rangle & \\mathbb{C} \\end{pmatrix}^\\text{T} ,

.. math::

   X =
   \\begin{pmatrix}
   1 & -\\mathbf{b}_1^\\text{T} & \\frac{1}{2} (\\mathbf{b}_1 \\otimes
   \\mathbf{b}_1) \\text{T} \\\\
   \\vdots & \\vdots & \\vdots \\\\
   1 & -\\mathbf{b}_n^\\text{T} & \\frac{1}{2} (\\mathbf{b}_n \\otimes
   \\mathbf{b}_n) ^\\text{T}
   \\end{pmatrix} ,

where $n$ is the number of acquisitions and $\\langle\\mathbf{D}\\rangle$,
$\\mathbb{C}$, $\\mathbf{b}_i$, and $(\\mathbf{b}_i \\otimes \\mathbf{b}_i)$
are represented by column vectors using Voigt notation. Estimation of the
model parameters requires that
$\\text{rank}(\\mathbf{X}^\\text{T}\\mathbf{X}) = 28$.
This can be achieved by combining acquisitions with b-tensors with different
shapes, sizes, and orientations.

For details, please see :footcite:p:`Westin2016`.

Usage example
=============

QTI can be fit to data using the module `dipy.reconst.qti`. Let's start by
importing the required modules and functions:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.qti as qti

###############################################################################
# As QTI requires data with tensor-valued encoding, let's load an example
# dataset acquired with q-space trajectory encoding (QTE):

fdata, fbvals, fbvecs, fmask = get_fnames(name="qte_lte_pte")
data, affine = load_nifti(fdata)
mask, _ = load_nifti(fmask)
bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
btens = np.array(["LTE" for i in range(61)] + ["PTE" for i in range(61)])
gtab = gradient_table(bvals, bvecs=bvecs, btens=btens)

###############################################################################
# The dataset contains 122 volumes of which the first half were acquired with
# linear tensor encoding (LTE) and the second half with planar tensor encoding
# (PTE). We can confirm this by calculating the ranks of the b-tensors in the
# gradient table.

ranks = np.array([np.linalg.matrix_rank(b) for b in gtab.btens])
for i, ell in enumerate(["b = 0", "LTE", "PTE"]):
    print(f"{np.sum(ranks == i)} volumes with {ell}")

###############################################################################
# Now that we have data acquired with tensor-valued diffusion encoding and the
# corresponding gradient table containing the b-tensors, we can fit QTI to the
# data as follows:

qtimodel = qti.QtiModel(gtab)
qtifit = qtimodel.fit(data, mask=mask)

###############################################################################
# QTI parameter maps can accessed as the attributes of `qtifit`. For instance,
# fractional anisotropy (FA) and microscopic fractional anisotropy (μFA) maps
# can be calculated as:

fa = qtifit.fa
ufa = qtifit.ufa

###############################################################################
# Finally, let's reproduce Figure 9 from :footcite:p:`Westin2016` to visualize
# more QTI parameter maps:

z = 36

fig, ax = plt.subplots(3, 4, figsize=(12, 9))

background = np.zeros(data.shape[0:2])  # Black background for figures
for i in range(3):
    for j in range(4):
        ax[i, j].imshow(background, cmap="gray")
        ax[i, j].set_xticks([])
        ax[i, j].set_yticks([])

ax[0, 0].imshow(np.rot90(qtifit.md[:, :, z]), cmap="gray", vmin=0, vmax=3e-3)
ax[0, 0].set_title("MD")
ax[0, 1].imshow(np.rot90(qtifit.v_md[:, :, z]), cmap="gray", vmin=0, vmax=0.5e-6)
ax[0, 1].set_title("V$_{MD}$")
ax[0, 2].imshow(np.rot90(qtifit.v_shear[:, :, z]), cmap="gray", vmin=0, vmax=0.5e-6)
ax[0, 2].set_title("V$_{shear}$")
ax[0, 3].imshow(np.rot90(qtifit.v_iso[:, :, z]), cmap="gray", vmin=0, vmax=1e-6)
ax[0, 3].set_title("V$_{iso}$")

ax[1, 0].imshow(np.rot90(qtifit.c_md[:, :, z]), cmap="gray", vmin=0, vmax=0.25)
ax[1, 0].set_title("C$_{MD}$")
ax[1, 1].imshow(np.rot90(qtifit.c_mu[:, :, z]), cmap="gray", vmin=0, vmax=1)
ax[1, 1].set_title("C$_{μ}$ = μFA$^2$")
ax[1, 2].imshow(np.rot90(qtifit.c_m[:, :, z]), cmap="gray", vmin=0, vmax=1)
ax[1, 2].set_title("C$_{M}$ = FA$^2$")
ax[1, 3].imshow(np.rot90(qtifit.c_c[:, :, z]), cmap="gray", vmin=0, vmax=1)
ax[1, 3].set_title("C$_{c}$")

ax[2, 0].imshow(np.rot90(qtifit.mk[:, :, z]), cmap="gray", vmin=0, vmax=1.5)
ax[2, 0].set_title("MK")
ax[2, 1].imshow(np.rot90(qtifit.k_bulk[:, :, z]), cmap="gray", vmin=0, vmax=1.5)
ax[2, 1].set_title("K$_{bulk}$")
ax[2, 2].imshow(np.rot90(qtifit.k_shear[:, :, z]), cmap="gray", vmin=0, vmax=1.5)
ax[2, 2].set_title("K$_{shear}$")
ax[2, 3].imshow(np.rot90(qtifit.k_mu[:, :, z]), cmap="gray", vmin=0, vmax=1.5)
ax[2, 3].set_title("K$_{μ}$")

fig.tight_layout()
plt.show()

###############################################################################
# For more information about QTI, please read the article by
# :footcite:t:`Westin2016`.
#
# References
# ----------
#
# .. footbibliography::
#



"""Testing tripwire module."""

from numpy.testing import assert_raises

from dipy.testing import assert_false, assert_true
from dipy.utils.tripwire import TripWire, TripWireError, is_tripwire


def test_is_tripwire():
    assert_false(is_tripwire(object()))
    assert_true(is_tripwire(TripWire("some message")))


def test_tripwire():
    # Test tripwire object
    silly_module_name = TripWire("We do not have silly_module_name")
    assert_raises(TripWireError, getattr, silly_module_name, "do_silly_thing")
    assert_raises(TripWireError, silly_module_name)
    # Check AttributeError can be checked too
    try:
        _ = silly_module_name.__wrapped__
    except TripWireError as err:
        assert_true(isinstance(err, AttributeError))
    else:
        raise RuntimeError("No error raised, but expected")



"""
==============================================
Reconstruct with Constant Solid Angle (Q-Ball)
==============================================

We show how to apply a Constant Solid Angle ODF (Q-Ball) model from
:footcite:t:`Aganj2010` to your datasets.

First import the necessary modules:
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.shm import CsaOdfModel
from dipy.segment.mask import median_otsu
from dipy.viz import actor, window

###############################################################################
# Download and read the data for this tutorial and load the raw diffusion data
# and the affine.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# img contains a nibabel Nifti1Image object (data) and gtab contains a
# GradientTable object (gradient information e.g. b-values). For example to
# read the b-values it is possible to write print(gtab.bvals).

print(f"data.shape {data.shape}")

###############################################################################
# Remove most of the background using DIPY's mask module.

maskdata, mask = median_otsu(
    data, vol_idx=range(10, 50), median_radius=3, numpass=1, autocrop=True, dilate=2
)

###############################################################################
# We instantiate our CSA model with spherical harmonic order ($l$) of 4

csamodel = CsaOdfModel(gtab, 4)

###############################################################################
# `Peaks_from_model` is used to calculate properties of the ODFs (Orientation
# Distribution Function) and return for
# example the peaks and their indices, or GFA which is similar to FA but for
# ODF based models. This function mainly needs a reconstruction model, the
# data and a sphere as input. The sphere is an object that represents the
# spherical discrete grid where the ODF values will be evaluated.

csapeaks = peaks_from_model(
    model=csamodel,
    data=maskdata,
    sphere=default_sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_odf=False,
    normalize_peaks=True,
)

GFA = csapeaks.gfa

print(f"GFA.shape {GFA.shape}")

###############################################################################
# Apart from GFA, csapeaks also has the attributes peak_values, peak_indices
# and ODF. peak_values shows the maxima values of the ODF and peak_indices
# gives us their position on the discrete sphere that was used to do the
# reconstruction of the ODF. In order to obtain the full ODF, return_odf
# should be True. Before enabling this option, make sure that you have enough
# memory.
#
# Let's visualize the ODFs of a small rectangular area in an axial slice of the
# splenium of the corpus callosum (CC).

data_small = maskdata[13:43, 44:74, 28:29]

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()

csaodfs = csamodel.fit(data_small).odf(default_sphere)

###############################################################################
# It is common with CSA ODFs to produce negative values, we can remove those
# using ``np.clip``

csaodfs = np.clip(csaodfs, 0, np.max(csaodfs, -1)[..., None])
csa_odfs_actor = actor.odf_slicer(
    csaodfs, sphere=default_sphere, colormap="plasma", scale=0.4
)
csa_odfs_actor.display(z=0)

scene.add(csa_odfs_actor)
print("Saving illustration as csa_odfs.png")
window.record(scene=scene, n_frames=1, out_path="csa_odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Constant Solid Angle ODFs.
#
#
# References
# ----------
#
# .. footbibliography::
#
